from copy import deepcopy
from enum import IntEnum
from pathlib import Path

from PyQt6.QtWidgets import QVBoxLayout, QPushButton, QFrame, QLabel, QHBoxLayout, QComboBox, \
    QCheckBox, QRadioButton, QButtonGroup

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.util.Helpers import Helpers
from otlmow_gui.GUI.screens.ExportData_elements.ExportAllDataSubScreen import ExportAllDataSubScreen
from otlmow_gui.GUI.screens.ExportData_elements.ExportFilteredDataSubScreen import ExportFilteredDataSubScreen
from otlmow_gui.GUI.screens.Screen import Screen


class ExportDataScreen(Screen):
    class ExportOptionId(IntEnum):
        ALL_DATA = 0
        ONLY_UNEDITED_DATA = 1

    def __init__(self, language_settings=None):
        super().__init__()
        self._ = language_settings
        self.container_insert_data_screen = QVBoxLayout()

        self.button_group = QButtonGroup()
        self.radio_button_export_all_data = QRadioButton()
        self.radio_button_export_only_unedited_data = QRadioButton()

        self.sub_screen_option_0_nothing = None
        self.sub_screen_option_1_all_data = None
        self.sub_screen_option_2_only_unedited_data = None

        self.export_btn = QPushButton()
        self.input_file_label = QLabel()
        self.extra_option_csv = QCheckBox()
        self.file_extension_selection = QComboBox()
        self.title = QLabel()
        self.file_type_label = QLabel()
        self.main_window = None
        self.relations_split_optionality = QCheckBox()
        self.supported_export_formats:dict = deepcopy(global_vars.supported_file_formats)
        if "SDF" in self.supported_export_formats:
            self.supported_export_formats.pop("SDF") # not yet supported for export in V0.5.0
        self.init_ui()

    def init_ui(self) -> None:
        """
        Initializes the user interface for the export data screen. This method sets up the layout
        by adding spacing, a menu, and configuring the container's margins and stretch properties
        to ensure proper alignment and appearance.

        :return: None
        """

        self.container_insert_data_screen.addSpacing(10)
        self.container_insert_data_screen.addWidget(self.create_menu())
        self.container_insert_data_screen.setContentsMargins(0, 0, 0, 0)
        # self.container_insert_data_screen.addStretch()
        self.setLayout(self.container_insert_data_screen)

    def create_menu(self) -> QFrame:
        """
         Creates and configures a menu for the export data screen.
         This method sets up the layout, adds various widgets including titles, options, and
         buttons, and returns the constructed menu as a QFrame.

        :return: A QFrame containing the configured menu for the export data screen.
        :rtype: QFrame
        """

        window = QFrame()
        window.setProperty('class', 'background-box')

        window_layout = QVBoxLayout()

        self.title.setText(self._('export_to_davie'))
        self.title.setProperty('class', 'sub-title')


        self.extra_option_csv.setText(self._('export assets in different files'))
        self.extra_option_csv.setHidden(True)

        window_layout.addWidget(self.title)
        window_layout.addWidget(self.create_radio_button_box())
        window_layout.addWidget(self.create_horizontal_line())

        # screen when the user hasn't selected any radiobutton
        self.sub_screen_option_0_nothing = Screen()
        sub_window_layout_nothing = QVBoxLayout()
        sub_window_layout_nothing.addStretch()
        self.sub_screen_option_0_nothing.setLayout(sub_window_layout_nothing)


        # screen when the user selects the first option to export all data
        self.sub_screen_option_1_all_data = ExportAllDataSubScreen(language_settings=self._)


        # screen when the user selects the second option to export only unedited data
        self.sub_screen_option_2_only_unedited_data = ExportFilteredDataSubScreen(language_settings=self._)
        sub_window_layout_nothing = QVBoxLayout()
        sub_window_layout_nothing.addStretch()
        self.sub_screen_option_2_only_unedited_data.setLayout(sub_window_layout_nothing)

        window_layout.addWidget(self.sub_screen_option_0_nothing)
        window_layout.addWidget(self.sub_screen_option_1_all_data)
        window_layout.addWidget(self.sub_screen_option_2_only_unedited_data)

        self.sub_screen_option_1_all_data.setHidden(True)
        self.sub_screen_option_2_only_unedited_data.setHidden(True)

        window.setLayout(window_layout)
        return window

    def create_horizontal_line(self):
        horizontal_line = QFrame()
        horizontal_line.setFrameShape(QFrame.Shape.HLine)
        horizontal_line.setLineWidth(1)
        horizontal_line.setMidLineWidth(2)
        horizontal_line.setFrameShadow(QFrame.Shadow.Sunken)
        return horizontal_line

    def create_radio_button_box(self) -> QFrame:


        button_box_frame = QFrame()
        button_box_layout = QHBoxLayout()

        self.radio_button_export_all_data.setText(self._('All data'))
        self.radio_button_export_only_unedited_data.setText(self._('Without unedited data'))

        self.button_group.addButton(self.radio_button_export_all_data,self.ExportOptionId.ALL_DATA)
        # self.button_group.setId(self.radio_button_export_all_data,
        #                         self.ExportOptionId.ALL_DATA)

        self.button_group.addButton(self.radio_button_export_only_unedited_data,self.ExportOptionId.ONLY_UNEDITED_DATA)
        # self.button_group.setId(self.radio_button_export_only_unedited_data,
        #                         self.ExportOptionId.ONLY_UNEDITED_DATA)

        self.button_group.idClicked.connect(self.on_radio_button_click)

        button_box_layout.addWidget(self.radio_button_export_all_data)
        button_box_layout.addStretch(1)
        button_box_layout.addWidget(self.radio_button_export_only_unedited_data)
        button_box_layout.addStretch(5)

        button_box_frame.setLayout(button_box_layout)


        return button_box_frame

    def reset_ui(self, _):
        """
        Creates a button box containing the export button for the export data screen. This method
        sets up the layout for the button, assigns its properties, and connects the button's click
        event to a method for opening the file picker.

        :return: A QFrame containing the button box with the export button.
        :rtype: QFrame
        """
        super().reset_ui(_)
        self._ = _
        self.title.setText(self._('export_to_davie'))

    def on_radio_button_click(self,id:int):
        if (id == self.ExportOptionId.ALL_DATA and self.sub_screen_option_1_all_data.isHidden()):
            if not self.sub_screen_option_0_nothing.isHidden():
                self.sub_screen_option_0_nothing.setHidden(True)
            if not self.sub_screen_option_2_only_unedited_data.isHidden():
                self.sub_screen_option_2_only_unedited_data.setHidden(True)
            self.sub_screen_option_1_all_data.setHidden(False)
        elif (id == self.ExportOptionId.ONLY_UNEDITED_DATA and self.sub_screen_option_2_only_unedited_data.isHidden()):
            if not self.sub_screen_option_0_nothing.isHidden():
                self.sub_screen_option_0_nothing.setHidden(True)
            if not self.sub_screen_option_1_all_data.isHidden():
                self.sub_screen_option_1_all_data.setHidden(True)
            self.sub_screen_option_2_only_unedited_data.setHidden(False)


    def open_folder_of_created_export_files(self, document_path:Path):
        Helpers.open_folder_and_select_document(document_path)



class FDOToolboxProcessError(Exception):
    def __init__(self, language, command: str, fdo_toolbox_error:str):
        self._ = language
        self.command = command
        self.fdo_toolbox_error = fdo_toolbox_error

        super().__init__()

    def __str__(self):
        # "An error occured during FDO toolbox call:  \nCall: {0} \nError:\n{1}"
        translation = self._("FDOToolbox_call_error_message")
        return translation.format(self.command, self.fdo_toolbox_error)



from typing import List, Optional

from PyQt6.QtWidgets import QVBoxLayout, QFrame, QHBoxLayout, QWidget, QSizePolicy
from otlmow_model.OtlmowModel.BaseClasses.RelationInteractor import RelationInteractor
from otlmow_model.OtlmowModel.Classes.ImplementatieElement.AIMObject import AIMObject
from otlmow_model.OtlmowModel.Classes.ImplementatieElement.RelatieObject import RelatieObject

from otlmow_gui.Domain.step_domain.RelationChangeDomain import RelationChangeDomain
from otlmow_gui.GUI.dialog_windows.DefineHeeftBetrokkeneRelationWindow import \
    DefineHeeftBetrokkeneRelationWindow
from otlmow_gui.GUI.screens.MapScreen import MapScreen
from otlmow_gui.GUI.screens.RelationChange_elements.ExistingRelationListWidget import \
    ExistingRelationListWidget
from otlmow_gui.GUI.screens.RelationChange_elements.ObjectListWidget import ObjectListWidget
from otlmow_gui.GUI.screens.RelationChange_elements.PossibleRelationListWidget import \
    PossibleRelationListWidget
from otlmow_gui.GUI.screens.RelationChange_elements.RelationChangeHelpers import RelationChangeHelpers
from otlmow_gui.GUI.screens.general_elements.ButtonWidget import ButtonWidget
from otlmow_gui.GUI.screens.screen_interface.RelationChangeScreenInterface import \
    RelationChangeScreenInterface
import qtawesome as qta

class RelationChangeScreen(RelationChangeScreenInterface):
    """
    Represents the screen for managing relation changes in the application.

    This class provides the user interface for displaying and modifying relations
    between objects, including options for selecting, expanding, and defining relations.
    It integrates various UI components to facilitate user interactions and data management.

    Args:
        language_settings (optional): Language settings for the user interface.

    Attributes:
        selected_object_col1 (optional): The currently selected object in the first column.
        container_insert_data_screen (QVBoxLayout): Layout for organizing UI components vertically.
        window (optional): Reference to the main window for the screen.
        window_layout (optional): Layout for the window.
        input_field (optional): Input field for user data entry.
        input_file_button (optional): Button for selecting input files.
        frame_layout (optional): Layout for organizing frame components.
        objects_list_gui (ObjectListWidget): GUI component for displaying object lists.
        possible_relation_list_gui (PossibleRelationListWidget): GUI component for displaying possible relations.
        existing_relation_list_gui (ExistingRelationListWidget): GUI component for displaying existing relations.
    """


    def __init__(self, language_settings=None):
        super().__init__()
        self._ = language_settings

        self.selected_object_col1 = None

        #gui elements
        self.container_insert_data_screen = QVBoxLayout()
        self.window = None
        self.window_layout = None

        self.input_field = None
        self.input_file_button = None

        self.frame_layout = None

        self.objects_list_gui = ObjectListWidget(self._,self)
        self.possible_relation_list_gui = PossibleRelationListWidget(self._,self)
        self.existing_relation_list_gui = ExistingRelationListWidget(self._,self)

        self.possible_relation_frame:QFrame = None
        self.existing_relation_frame:QFrame = None

        self.map_window:MapScreen = None

        self.init_ui()




    def set_gui_lists_to_loading_state(self) -> None:
        """
        Sets the GUI lists to a loading state.

        This method clears the current contents of the objects and relations lists
        in the user interface and adds a loading placeholder to indicate that data
        is being processed. It provides visual feedback to the user while the
        application retrieves or updates data.

        :param self: The instance of the class.
        :returns: None
        """

        self.objects_list_gui.clear()
        self.possible_relation_list_gui.clear()
        self.existing_relation_list_gui.clear()

        self.objects_list_gui.add_loading_placeholder()
        self.possible_relation_list_gui.add_loading_placeholder()
        self.existing_relation_list_gui.add_loading_placeholder()


    def paintEvent(self, a0):
        self.synchronize_subtext_label_heights()
        super().paintEvent(a0)


    def synchronize_subtext_label_heights(self) -> None:
        """
        Synchronizes the heights of subtext labels across different GUI lists.

        This method adjusts the minimum height of the subtext labels in the objects
        list and existing relations list to match the height of the subtext label in
        the possible relations list. This ensures a consistent appearance in the user
        interface.

        :param self: The instance of the class.
        :returns: None
        """

        possible_subtext_frame = self.possible_relation_list_gui.get_adjustable_subtext_frame()
        object_subtext_frame = self.objects_list_gui.get_adjustable_subtext_frame()
        exist_subtext_frame = self.existing_relation_list_gui.get_adjustable_subtext_frame()

        frame_rect_height = possible_subtext_frame.frameRect().height()
        frame_rect_object = object_subtext_frame.frameRect()
        frame_rect_exist = exist_subtext_frame.frameRect()

        if self.existing_relation_frame.isVisible():
            if frame_rect_height != frame_rect_object.height():
                object_subtext_frame.setMinimumHeight(frame_rect_height)
            if frame_rect_height != frame_rect_exist.height():
                exist_subtext_frame.setMinimumHeight(frame_rect_height)
        else:
            object_subtext_frame.setMinimumHeight(15)

    def init_ui(self) -> None:
        """
        Initializes the user interface for the relation change screen.

        This method sets up the main layout by adding the menu to the container
        and configuring the layout margins. It ensures that the user interface
        is properly structured for user interaction.

        :param self: The instance of the class.
        :returns: None
        """

        self.container_insert_data_screen.addWidget(self.create_menu())
        self.container_insert_data_screen.setContentsMargins(0, 0, 0, 0)

        self.setLayout(self.container_insert_data_screen)


    def create_menu(self) -> QWidget:
        """
        Creates and configures the menu for the relation change screen.

        This method initializes a QWidget that serves as the menu container,
        setting its layout and adding the necessary components. It ensures that
        the menu is properly structured and styled for the user interface.

        :param self: The instance of the class.
        :returns: The configured menu widget.
        :rtype: QWidget
        """

        self.window = QWidget()
        self.window.setProperty('class', 'background-box')

        self.window_layout = QVBoxLayout()
        self.window_layout.setContentsMargins(0,0,0,0)
        self.window_layout.addWidget(self.horizontal_layout())

        self.window.setLayout(self.window_layout)

        return self.window


    def fill_object_list(self, objects: List[RelationInteractor]) -> None:
        """
        Fills the object list GUI with the provided AIM objects.

        This method updates the objects list in the user interface by passing the
        specified list of AIM objects to the GUI component responsible for displaying
        the objects. It ensures that the list is populated with the current data for
        user interaction.

        :param self: The instance of the class.
        :param objects: A list of AIM objects to be displayed in the object list.
        :type objects: List[AIMObject]
        :returns: None
        """

        self.objects_list_gui.fill_list(source_object=None, objects=objects, last_added=[])


    def fill_existing_relations_list(self, relations_objects: List[RelatieObject],
                                     last_added: list[RelatieObject] = None) -> None:
        """
        Fills the existing relations list GUI with the provided relation objects.

        This method updates the existing relations list in the user interface by
        populating it with the specified relation objects and any last added relations.
        It ensures that the list reflects the current state of relations for user interaction.

        :param self: The instance of the class.
        :param relations_objects: A list of relation objects to be displayed in the existing relations list.
        :type relations_objects: List[RelatieObject]
        :param last_added: An optional list of last added relation objects to highlight in the list.
        :type last_added: list[RelatieObject], optional
        :returns: None
        """

        if last_added is None:
            last_added = []
        self.existing_relation_list_gui.fill_list(source_object=None,
                                                  objects=relations_objects,
                                                  last_added=last_added)


    def fill_possible_relations_list(self, source_object: Optional[RelationInteractor],
                                     relations: dict[str, list[RelatieObject]],
                                     last_added=None) -> None:
        """
        Fills the possible relations list GUI with relations for a given source object.

        This method updates the possible relations list in the user interface by
        populating it with relations associated with the specified source object.
        It also highlights any previously added relations to provide context for the user.

        :param self: The instance of the class.
        :param source_object: The source object for which possible relations are being displayed.
        :type source_object: AIMObject
        :param relations: A dictionary mapping relation types to lists of relation objects.
        :type relations: dict[str, list[RelatieObject]]
        :param last_added: An optional list of last added relation objects to highlight in the list.
        :type last_added: list[RelatieObject], optional
        :returns: None
        """

        if last_added is None:
            last_added = []
        self.possible_relation_list_gui.fill_list(source_object=source_object,
                                                  objects=relations,
                                                  last_added=last_added)


    def horizontal_layout(self) -> QFrame:
        """
        Creates a horizontal layout for displaying object and relation lists.

        This method constructs a QFrame containing a horizontal layout that includes
        three object list GUI components: the objects list, possible relations list,
        and existing relations list. It sets the layout's spacing and stretch factors
        to ensure a balanced and responsive design.

        :param self: The instance of the class.
        :returns: The configured frame containing the horizontal layout.
        :rtype: QFrame
        """

        frame = QFrame()
        self.frame_layout = QHBoxLayout()
        self.frame_layout.setSpacing(0)

        self.frame_layout.addWidget(self.objects_list_gui.create_object_list_gui())

        self.possible_relation_frame =self.possible_relation_list_gui.create_object_list_gui(multi_select=True)
        self.frame_layout.addWidget(self.possible_relation_frame)

        self.existing_relation_frame = self.existing_relation_list_gui.create_object_list_gui(multi_select=True)
        self.frame_layout.addWidget( self.existing_relation_frame)

        map_button = ButtonWidget()
        map_button.setIcon(qta.icon("mdi.map",color='white',options=[{
                'scale_factor': 1.5
            }]))
        map_button.setSizePolicy(QSizePolicy.Policy.Minimum,QSizePolicy.Policy.Expanding)
        map_button.setProperty('class', 'map-button')
        map_button.clicked.connect(lambda: self.toggle_map_window())
        map_button.setToolTip(self._("Toon kaart met OTL-assets uit de eerste kolom"))
        # map_button.setContentsMargins(22,0,0,0)

        self.frame_layout.addWidget(map_button)
        # self.step3_map: MapScreen = MapScreen(self._)

        self.frame_layout.setStretch(0, 5)
        self.frame_layout.setStretch(1, 5)
        self.frame_layout.setStretch(2, 5)
        # self.frame_layout.setStretch(3, 1)

        # self.possible_relation_frame.setHidden(True)
        # self.existing_relation_frame.setHidden(True)

        frame.setLayout(self.frame_layout)
        return frame


    def reset_ui(self, _):
        super().reset_ui(_)
        self._ = _

    def fill_object_attribute_field(self, object_attribute_dict: dict) -> None:
        """
        Fills the object attribute field with the provided attributes.

        This method updates the object attribute field in the user interface by
        passing a dictionary of object attributes to the corresponding GUI component.
        It ensures that the displayed attributes are current and reflect the selected object.

        :param self: The instance of the class.
        :param object_attribute_dict: A dictionary containing object attributes to be displayed.
        :type object_attribute_dict: dict
        :returns: None
        """

        self.objects_list_gui.fill_object_attribute_field(
            object_attribute_dict=object_attribute_dict)

    def fill_possible_relation_attribute_field(self,
                                               possible_relation_attribute_dict: dict) -> None:
        """
        Fills the possible relation attribute field with the provided attributes.

        This method updates the possible relation attribute field in the user interface
        by passing a dictionary of possible relation attributes to the corresponding GUI
        component. It ensures that the displayed attributes accurately reflect the selected
        possible relation.

        :param self: The instance of the class.
        :param possible_relation_attribute_dict: A dictionary containing possible relation attributes to be displayed.
        :type possible_relation_attribute_dict: dict
        :returns: None
        """

        self.possible_relation_list_gui.fill_object_attribute_field(
            object_attribute_dict=possible_relation_attribute_dict)

    def fill_existing_relation_attribute_field(self,
                                               existing_relation_attribute_dict: dict) -> None:
        """
        Fills the existing relation attribute field with the provided attributes.

        This method updates the existing relation attribute field in the user interface
        by passing a dictionary of existing relation attributes to the corresponding GUI
        component. It ensures that the displayed attributes accurately reflect the selected
        existing relation.

        :param self: The instance of the class.
        :param existing_relation_attribute_dict: A dictionary containing existing relation attributes to be displayed.
        :type existing_relation_attribute_dict: dict
        :returns: None
        """

        self.existing_relation_list_gui.fill_object_attribute_field(
            object_attribute_dict=existing_relation_attribute_dict)

    def expand_existing_relations_folder_of(self, relation_typeURI: str) -> None:
        """Expands the folder of existing relations for a specified relation type.

        This method determines if the given relation type URI is unique across namespaces
        and retrieves its abbreviated form. It then expands the corresponding folder
        in the existing relations list GUI to display related items.

        :param self: The instance of the class.
        :param relation_typeURI: The type URI of the relation whose folder is to be expanded.
        :type relation_typeURI: str
        :returns: None
        """

        add_namespace = RelationChangeHelpers.is_unique_across_namespaces(
            typeURI=relation_typeURI,
            objects=RelationChangeDomain.shown_objects)
        abbr_relation_typeURI = RelationChangeHelpers.get_abbreviated_typeURI(
            typeURI=relation_typeURI,
            add_namespace=add_namespace,
            is_relation=True)
        self.existing_relation_list_gui.expand_folder_of(typeURI=abbr_relation_typeURI)

    def expand_possible_relations_folder_of(self, relation_typeURI: str) -> None:
        """Expands the folder of possible relations for a specified relation type.

        This method checks if the given relation type URI is unique across namespaces
        and retrieves its abbreviated form. It then expands the corresponding folder
        in the possible relations list GUI to display related items.

        :param self: The instance of the class.
        :param relation_typeURI: The type URI of the relation whose folder is to be expanded.
        :type relation_typeURI: str
        :returns: None
        """

        add_namespace = RelationChangeHelpers.is_unique_across_namespaces(
            typeURI=relation_typeURI,
            objects=RelationChangeDomain.shown_objects)
        abbr_relation_typeURI = RelationChangeHelpers.get_abbreviated_typeURI(
            typeURI=relation_typeURI,
            add_namespace=add_namespace,
            is_relation=True)
        self.possible_relation_list_gui.expand_folder_of(abbr_relation_typeURI)

    def showMultiSelectionHeeftBetrokkeneAttributeDialogWindow(self,
                                                               data_list_and_relation_objects: list) -> None:
        """Displays a dialog window for defining the 'HeeftBetrokkene' relation.

        This method initializes and opens a dialog window that allows the user to
        define the 'HeeftBetrokkene' relation based on the provided data list and
        relation objects. It ensures that the user can interactively set the necessary
        attributes for the relation.

        :param self: The instance of the class.
        :param data_list_and_relation_objects: A list containing data and relation objects
                                                relevant to the 'Heeft Betrokkene' relation.
        :type data_list_and_relation_objects: list
        :returns: None
        """

        dialogWindow = DefineHeeftBetrokkeneRelationWindow(
            language_settings=self._,
            data_list_and_relation_objects=data_list_and_relation_objects)
        dialogWindow.draw_define_heeft_betrokkene_rol_window()

    def clear_possible_relation_elements(self) -> None:
        """
        Clears all possible relation elements from the user interface.
        This method resets the displayed lists and fields related to possible relations,
        ensuring that no outdated or irrelevant information is shown.

        :return: None
        """

        self.fill_possible_relations_list(None, {})
        self.fill_object_attribute_field({})
        self.fill_possible_relation_attribute_field({})

    def update_color_scheme(self):
        self.objects_list_gui.update_color_scheme()
        self.possible_relation_list_gui.update_color_scheme()
        self.existing_relation_list_gui.update_color_scheme()

    def set_object_search_bar_text(self, search_text: str) -> None:
        """
        Sets the text of the object search bar in the GUI.

        This method updates the search bar with the provided search text, allowing users to filter the displayed objects based on their input.

        :param search_text: The text to set in the object search bar.
        :type search_text: str

        :return: None
        """
        self.objects_list_gui.set_search_text(search_text)

    def set_possible_relation_search_bar_text(self, search_text: str) -> None:
        """
        Sets the text of the possible relation search bar in the GUI.

        This method updates the search bar with the provided search text, enabling users to filter the displayed possible relations based on their input.

        :param search_text: The text to set in the possible relation search bar.
        :type search_text: str

        :return: None
        """
        self.possible_relation_list_gui.set_search_text(search_text)

    def set_existing_relation_search_bar_text(self, search_text: str) -> None:
        """
        Sets the text of the existing relation search bar in the GUI.

        This method updates the search bar with the provided search text, allowing users to filter the displayed existing relations based on their input.

        :param search_text: The text to set in the existing relation search bar.
        :type search_text: str

        :return: None
        """
        self.existing_relation_list_gui.set_search_text(search_text)

    def get_current_object_list_content_dict(self) -> dict[str,list]:
        return self.objects_list_gui.get_current_list_content_dict()

    def set_selected_object(self, identificator:str):
        self.objects_list_gui.select_item_via_identificator(identificator)

    def toggle_map_window(self):
        if self.map_window:
            self.map_window.close()
        else:
            self.map_window = MapScreen(self._,parent_screen=self)
            self.map_window.show()
            self.map_window.start_async_reload()

    def is_show_all_OTL_relations_checked(self) -> bool:
        return self.possible_relation_list_gui.is_show_all_OTL_relations_checked()


import json
import re
from pathlib import Path
from typing import List

import qtawesome as qta

from PyQt6.QtWebEngineCore import QWebEngineSettings
from PyQt6.QtWidgets import QVBoxLayout, QLabel, QComboBox, QSizePolicy, \
    QWidget, QFrame, QHBoxLayout, QCheckBox, QSlider
from PyQt6.QtWebEngineWidgets import QWebEngineView
from PyQt6.QtWebChannel import QWebChannel
from PyQt6.QtCore import QObject, pyqtSlot, QUrl, Qt
from otlmow_model.OtlmowModel.BaseClasses.OTLObject import OTLObject
from otlmow_visuals.PyVisWrapper import PyVisWrapper

from otlmow_gui.Domain import global_vars

from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.step_domain.RelationChangeDomain import RelationChangeDomain
from otlmow_gui.Domain.util.Helpers import Helpers
from otlmow_gui.GUI.dialog_windows.NotificationWindow import NotificationWindow
from otlmow_gui.GUI.dialog_windows.OverwriteGraphWarningWindow import OverwriteGraphWarningWindow
from otlmow_gui.GUI.screens.DataVisualisation_elements.VisualisationHelper import VisualisationHelper
from otlmow_gui.GUI.screens.Screen import Screen
from otlmow_gui.GUI.screens.general_elements.ButtonWidget import ButtonWidget
from otlmow_gui.GUI.translation.GlobalTranslate import GlobalTranslate


class Backend(QObject):

    def __init__(self, parent_screen):
        super().__init__()
        self.parent_screen = parent_screen

    @pyqtSlot(str)
    def receive_new_combined_data(self, message):
        """Receives receive_new_combined_data from JavaScript."""
        data = json.loads(message)  # Convert string back to dict

        OTLLogger.logger.debug(f"Received new combined data: {message}")

        # test_json_path = global_vars.current_project.get_current_visuals_folder_path() / "new_nodes.json"
        # with open(test_json_path, 'w') as json_file:
        #     json.dump(data, json_file)


        new_node_data_str = json.dumps(data['nodeList']) #nodes and their position (including edgeJointNodes)
        new_edge_data_str = json.dumps(data['edgeList']) #edges (including subEdges)
        new_relationIdToSubEdges_data_str = json.dumps(data['relationIdToSubEdgesList']) #data supporting dynamic removal functionality
        new_relationIdToTotalSubEdgeCountList_data_str = json.dumps(data['relationIdToTotalSubEdgeCountList']) #data supporting dynamic removal functionality
        new_relationIdToJointNodesList_data_str = json.dumps(data['relationIdToJointNodesList']) #data supporting dynamic removal functionality
        new_SubEdgesToOriginalRelationIdList_data_str = json.dumps(data['SubEdgesToOriginalRelationIdList']) #data supporting dynamic removal functionality
        new_edgeJointNodesIdToConnectionDataDictList_data_str = json.dumps(data[
                                                                       'edgeJointNodesIdToConnectionDataDictList'])  # data supporting right-click edgeJointNode removal functionality
        new_collection_id_to_list_of_relation_ids_data_str = "{}"
        if 'collection_id_to_list_of_relation_ids' in data.keys():
            new_collection_id_to_list_of_relation_ids_data_str = json.dumps(data[
                                                                                   'collection_id_to_list_of_relation_ids'])  # data supporting hover to show collection edge functionality

        OTLLogger.logger.debug(new_collection_id_to_list_of_relation_ids_data_str)

        new_node_data_str = self.correct_node_title_attributes(new_node_data_str)

        try:
            html_path = self.parent_screen.get_current_html_path()
        except:
            OTLLogger.logger.error("Couldn't save the html because the html is outdated and the project is already closed")
            notification = NotificationWindow(title=GlobalTranslate._("Couldn't save visualisation"),
                                              message=GlobalTranslate._("Couldn't save the html because the html is outdated and the project is already closed\nTo fix this refresh the visualisation."))
            notification.exec()
            return

        self.parent_screen.save_html(file_path=html_path,
                                     new_node_data=new_node_data_str,
                                     new_edge_data=new_edge_data_str,
                                     new_relationIdToSubEdges_data=new_relationIdToSubEdges_data_str,
                                     new_relationIdToTotalSubEdgeCount_data=new_relationIdToTotalSubEdgeCountList_data_str,
                                     new_relationIdToJointNodes_data=new_relationIdToJointNodesList_data_str,
                                     new_SubEdgesToOriginalRelationId_data=new_SubEdgesToOriginalRelationIdList_data_str,
                                     new_edgeJointNodesIdToConnectionDataDict_data=new_edgeJointNodesIdToConnectionDataDictList_data_str,
                                     new_collection_id_to_list_of_relation_ids_dict=new_collection_id_to_list_of_relation_ids_data_str)
        global_vars.current_project.visualisation_uptodate.reset_full_state()
        global_vars.current_project.save_visualisation_python_support_data(self.parent_screen.std_vis_wrap)
    @pyqtSlot()
    def receive_network_changed_notification(self):
        self.parent_screen.set_graph_saved_status(False)

    @pyqtSlot()
    def receive_network_loaded_notification(self):
        OTLLogger.logger.info("js says network is loaded")
        self.parent_screen.js_is_loading = False
        self.parent_screen.loading_graph_status_label.setText("")

    @classmethod
    def correct_node_title_attributes(cls, new_node_data_str):
        # correcting the formatting of the title attribute of the node
        new_node_data_str = new_node_data_str.replace('"htmlTitle(\\', 'htmlTitle(')
        new_node_data_str = new_node_data_str.replace('</div>\\")"', '</div>")')
        return new_node_data_str




class DynamicDataVisualisationScreen(Screen):

    def __init__(self, _):
        super().__init__()
        # self.setWindowTitle("PyQt6 QWebEngine QWebChannel Example")
        # self.resize(800, 600)

        # data variables
        self.objects_in_memory = []
        self.std_vis_wrap = None
        self.js_is_loading = False
        self.max_slider_value = 50

        # translation model
        self._ = _

        #smaller support widgets
        self.refresh_needed_label = QLabel()
        self.graph_saved_status_label = QLabel()
        self.loading_graph_status_label = QLabel()
        self.visualisation_mode = QComboBox()
        self.collection_threshold_slider = QSlider(Qt.Orientation.Horizontal)
        self.collection_threshold_label = QLabel()
        self.too_many_objects_message = QLabel()
        self.relation_color_legend_title = QLabel()
        self.relation_id_to_relation_show_checkbox_dict = {}

        self.main_layout = QVBoxLayout()
        self.main_widget = self.create_main_widget()
        self.main_layout.addWidget(self.main_widget)
        self.setLayout(self.main_layout)



    def create_main_widget(self):
        main_widget = QWidget()
        main_widget.setProperty('class', 'background-box')

        window_layout = QVBoxLayout()
        window_layout.setContentsMargins(0, 0, 0, 0)

        # Set up the channel and backend for communication from js to python backend
        self.channel = QWebChannel()
        self.backend = Backend(parent_screen=self)
        self.channel.registerObject("backend", self.backend)

        # main QWebEngineView widget
        self.view = QWebEngineView(self)
        self.view.page().setWebChannel(self.channel)
        self.view.settings().setAttribute(QWebEngineSettings.WebAttribute.ShowScrollBars, False)
        self.view.setContentsMargins(0, 0, 0, 0)
        self.view.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)


        self.view.settings().setAttribute( QWebEngineSettings.WebAttribute.LocalContentCanAccessFileUrls, True)
        self.view.settings().setAttribute(
            QWebEngineSettings.WebAttribute.LocalContentCanAccessRemoteUrls, True)

        self.too_many_objects_message.setVisible(False)

        self.relation_color_legend_title.setText(self._("relations legend") + ":")

        window_layout.addWidget(self.create_button_container())
        window_layout.addWidget(self.view, 2)
        window_layout.addWidget(self.too_many_objects_message, 2)
        window_layout.addWidget(self.relation_color_legend_title)
        window_layout.addWidget(self.create_color_legend())

        main_widget.setLayout(window_layout)
        return main_widget

    def create_button_container(self):
        frame = QFrame()
        frame.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        frame_layout = QHBoxLayout()


        regenerate_btn = ButtonWidget()
        regenerate_btn.setIcon(qta.icon('mdi.refresh', color='white'))
        regenerate_btn.setProperty('class', 'primary-button')
        regenerate_btn.clicked.connect(lambda: OverwriteGraphWarningWindow(self,self._))
        regenerate_btn.setToolTip(self._("Regenerate the visualisation"))

        self.visualisation_mode.addItems(["1 Hiërarchische visualisatie",
                                          "2 Spinnenweb visualisatie",
                                          "3 Shell visualisatie",
                                          "4 ForceAtlas2Based visualisatie",
                                          ])
        self.visualisation_mode.setToolTip(self._("Options to generate the visualisation differently"))
        self.visualisation_mode.currentTextChanged.connect(self.on_visualisation_option_change)

        collection_threshold_tooltip = self._("Deze slider duidt de hoeveelheid van dezelfde relaties naar één asset nodig voor het bundelen in een collectie")
        self.collection_threshold_slider.setRange(0, 50)
        self.collection_threshold_slider.setValue(10)
        self.collection_threshold_slider.setToolTip(collection_threshold_tooltip)
        self.collection_threshold_slider.setFixedWidth(100)

        self.collection_threshold_label.setText((str(self.collection_threshold_slider.value())))
        self.collection_threshold_label.setToolTip(collection_threshold_tooltip)

        # Connect slider change to label update
        self.collection_threshold_slider.valueChanged.connect(self.on_value_changed)

        self.refresh_needed_label.setText(
            self._("The visualisation is outdated, refresh to see new changes"))
        self.refresh_needed_label.setStyleSheet('color:#DD1111;')
        # self.set_refresh_need_label_hidden_status(True)

        self.set_graph_saved_status(True)

        save_btn = ButtonWidget()
        # see all qta mdi icon options in: https://cdn.jsdelivr.net/npm/@mdi/font@5.9.55/preview.html
        save_btn.setIcon(qta.icon('mdi.content-save', color='white'))
        save_btn.setProperty('class', 'primary-button')
        save_btn.clicked.connect(lambda: self.save_in_memory_changes_to_html())
        save_btn.setToolTip(self._("Save changes"))

        help_btn = ButtonWidget()
        # see all qta mdi icon options in: https://cdn.jsdelivr.net/npm/@mdi/font@5.9.55/preview.html
        help_btn.setIcon(qta.icon('mdi.help', color='white'))
        help_btn.setProperty('class', 'primary-button')
        help_btn.clicked.connect(lambda: self.show_help_dialog_window())
        help_btn.setToolTip(self._("Help for data visualisation screen"))


        open_html_location_btn = ButtonWidget()
        # see all qta mdi icon options in: https://cdn.jsdelivr.net/npm/@mdi/font@5.9.55/preview.html
        open_html_location_btn.setIcon(qta.icon('mdi.folder-marker', color='white'))
        open_html_location_btn.setProperty('class', 'primary-button')
        open_html_location_btn.clicked.connect(lambda: Helpers.open_folder_and_select_document(
            global_vars.current_project.get_current_visuals_html_path()))
        open_html_location_btn.setToolTip(self._("Open visualisation html file location"))

        frame_layout.addWidget(regenerate_btn)
        frame_layout.addWidget(self.visualisation_mode)
        frame_layout.addWidget(self.collection_threshold_slider)
        frame_layout.addWidget(self.collection_threshold_label)
        frame_layout.addWidget(self.refresh_needed_label)
        frame_layout.addStretch()
        frame_layout.addWidget(self.loading_graph_status_label)
        frame_layout.addStretch()
        frame_layout.addWidget(self.graph_saved_status_label)
        frame_layout.addWidget(save_btn)
        frame_layout.addWidget(open_html_location_btn)
        frame_layout.addWidget(help_btn)
        frame_layout.setContentsMargins(0, 0, 0, 0)

        frame.setLayout(frame_layout)
        return frame

    def on_visualisation_option_change(self):
        global_vars.current_project.visualisation_uptodate.set_clear_all(True)
        self.update_refresh_need_label_state()

    def on_value_changed(self, val: int):
        self.collection_threshold_label.setText(str(val))
        self.check_if_slider_is_at_maxSliderValue()
        global_vars.current_project.visualisation_uptodate.set_clear_all(True)
        self.update_refresh_need_label_state()

    def show_help_dialog_window(self):
        title = self._("Help for data visualisation screen")
        message = self._("""
Help voor het gebruik van het datavisualisatie scherm:
        
- Linker-muisknop inhouden op OTL-asset => slepen
- Linker-muisknop op relatie pijl => een versleepbaar gewricht toevoegen
- Rechter-muisknop inhouden en slepen in lege ruimte => selectie box om meerder OTL-assets (+ gewrichten) te selecteren
- Ctrl-knop + Linker-muisknop op OTL-asset => toevoegen/verwijderen van OTL-asset aan selectie
        """)

        dialog = NotificationWindow(message = message,title=title)
        dialog.exec()

    def create_color_legend(self):

        frame = QFrame()
        self.frame_layout_legend = QHBoxLayout()
        self.frame_layout_legend.setContentsMargins(0, 0, 0, 0)
        self.fill_frame_layout_legend()
        frame.setLayout(self.frame_layout_legend)
        return frame

    def fill_frame_layout_legend(self):
        typeURIs_in_memory = [object_in_memory.typeURI for object_in_memory in
                              self.objects_in_memory]

        initial_state_per_relation_dict = {}
        key_list = list(self.relation_id_to_relation_show_checkbox_dict.keys())
        for relatie in typeURIs_in_memory:
            if relatie in key_list:
                initial_state_per_relation_dict[relatie] = self.relation_id_to_relation_show_checkbox_dict[
                    relatie].isChecked()
            else:
                initial_state_per_relation_dict[relatie] =  True


        for i in reversed(range(self.frame_layout_legend.count())):
            item = self.frame_layout_legend.takeAt(i)
            if item.widget():
                item.widget().deleteLater()
        self.relation_id_to_relation_show_checkbox_dict.clear()

        relatie_color_dict = PyVisWrapper().relatie_color_dict


        for relatie, color in relatie_color_dict.items():

            if relatie not in typeURIs_in_memory:
                continue

            item_frame = QFrame()
            item_layout_legend = QHBoxLayout()

            show_checkbox = QCheckBox()
            show_checkbox.setChecked(initial_state_per_relation_dict[relatie])
            show_checkbox.stateChanged.connect(self.create_on_stateChange_listener(relation_color=color))
            # store checkbox tied to it's relation typeURI
            self.relation_id_to_relation_show_checkbox_dict[
                relatie] = show_checkbox

            label = QLabel()
            relatie_name = relatie.split('#')[-1]
            label.setText(relatie_name)
            label.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Expanding)
            # label.setFixedWidth(50)

            color_label = QLabel()
            color_label.setStyleSheet(f"background-color: #{color}; ")
            color_label.setSizePolicy(QSizePolicy.Policy.Maximum,QSizePolicy.Policy.Expanding)
            color_label.setFixedWidth(30)




            item_layout_legend.addWidget(label)
            item_layout_legend.addWidget(show_checkbox)
            item_layout_legend.addWidget(color_label)
            item_layout_legend.setContentsMargins(10, 0, 0, 0)

            item_frame.setLayout(item_layout_legend)
            item_frame.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
            self.frame_layout_legend.addWidget(item_frame)

        if self.frame_layout_legend:
            self.frame_layout_legend.addStretch()


    def create_on_stateChange_listener(self,relation_color):
        return lambda state : self.set_relations_visibility(relation_color,state)

    def set_relations_visibility(self,relation_color:str, visibility:bool):
        if visibility:
            js_bool_hidden = "false"
        else:
            js_bool_hidden = "true"
        js_code = f"UpdateAllRelationHiddenStatesOfColor('{relation_color}',{js_bool_hidden})"
        OTLLogger.logger.debug(js_code)
        self.view.page().runJavaScript(js_code)

    def reset_ui(self, _):
        pass

    def load_html(self, html_loc):

        self.loading_graph_status_label.setText(self._("Loading graph"))

        self.view.setUrl(QUrl.fromLocalFile(str(html_loc).replace("\\","/")))

        if not self.std_vis_wrap:
            self.std_vis_wrap = VisualisationHelper.get_std_vis_wrap_instance()

        global_vars.current_project.load_visualisation_python_support_data(self.std_vis_wrap)

        self.set_graph_saved_status(True)

    def recreate_html(self,html_path:Path):
        OTLLogger.logger.debug(
            f"Executing DataVisualisationScreen.reload_html() for project {global_vars.current_project.eigen_referentie}",
            extra={"timing_ref": f"reload_html_{global_vars.current_project.eigen_referentie}"})

        self.objects_in_memory = self.load_assets()

        self.fill_frame_layout_legend()


        global_vars.current_project.visualisation_uptodate.reset_full_state()
        self.update_refresh_need_label_state()


        object_count = len(self.objects_in_memory)
        if object_count > VisualisationHelper.object_count_limit:
            # OBSOLETE: since V1.1 this will nearly never be entered since the threshold is so high
            # do not draw the visualisation when there are too many assets
            self.view.setVisible(False)
            if not self.too_many_objects_message.isVisible():
                self.too_many_objects_message.setVisible(True)

            translation = self._(
                "There are too many assets and relations to create a visualisation.\nmaximum: {0}\ncurrent: {1}")
            self.too_many_objects_message.setText(translation.format(
                VisualisationHelper.object_count_limit,
                object_count))

        else:

            self.too_many_objects_message.setVisible(False)
            if not self.view.isVisible():
                self.view.setVisible(True)

            self.std_vis_wrap = VisualisationHelper.create_html(html_loc= html_path,
                                                                objects_in_memory=self.objects_in_memory,
                                                                vis_mode=self.visualisation_mode.currentText(),
                                                                collection_threshold = self.collection_threshold_slider.value())
            global_vars.current_project.visualisation_uptodate.reset_full_state()
            global_vars.current_project.save_visualisation_python_support_data(self.std_vis_wrap)
            self.load_html(html_path)
            self.set_graph_saved_status(False)

        #reset the state of all color legend checkboxes
        self.relation_id_to_relation_show_checkbox_dict.clear()
        self.update_only_legend()

        object_count = len(self.objects_in_memory)
        OTLLogger.logger.debug(
            f"Executing DataVisualisationScreen.reload_html() for project {global_vars.current_project.eigen_referentie} ({object_count} objects)",
            extra={"timing_ref": f"reload_html_{global_vars.current_project.eigen_referentie}"})



    def get_current_html_path(self):
        return global_vars.current_project.get_current_visuals_html_path()

    def opened(self):
        if self.changed_project_bool:
            self.changed_project()


        elif not RelationChangeDomain.is_visualisation_uptodate():
            # if self.visualisation_mode.currentText() != "standaard visualisatie":
            #     self.set_refresh_need_label_hidden_status(False)
            #     return


            to_add_list = global_vars.current_project.visualisation_uptodate.get_to_be_inserted_relations()
            to_remove_list = global_vars.current_project.visualisation_uptodate.get_to_be_removed_relations()

            VisualisationHelper.add_new_relations(to_add_list=to_add_list,
                                   vis_wrap=self.std_vis_wrap, webview=self.view,relation_visible_dict=self.relation_id_to_relation_show_checkbox_dict)
            VisualisationHelper.remove_relations(to_remove_list=to_remove_list,
                                  vis_wrap=self.std_vis_wrap, webview=self.view)

            global_vars.current_project.visualisation_uptodate.reset_relations_uptodate()

        self.update_refresh_need_label_state()

    def update_refresh_need_label_state(self):
        if global_vars.current_project.visualisation_uptodate.is_uptodate():
            self.set_refresh_need_label_hidden_status(True)
        else:
            self.set_refresh_need_label_hidden_status(False)

    def update_slider_range(self):
        asset_count_in_project = len(RelationChangeDomain.get_shown_objects())
        self.max_slider_value = asset_count_in_project + 1
        self.collection_threshold_slider.blockSignals(True)
        self.collection_threshold_slider.setRange(0, self.max_slider_value)
        if self.max_slider_value > 10:
            self.collection_threshold_slider.setValue(10)
            self.collection_threshold_label.setText(str(self.collection_threshold_slider.value()))
        else:
            self.collection_threshold_slider.setValue(self.max_slider_value)
            self.check_if_slider_is_at_maxSliderValue()
        self.collection_threshold_slider.blockSignals(False)

    def check_if_slider_is_at_maxSliderValue(self):
        if self.collection_threshold_slider.value() >= self.max_slider_value:
            self.collection_threshold_label.setText(self._("uit"))

    def update_only_legend(self):
        self.objects_in_memory = self.load_assets()
        self.fill_frame_layout_legend()

    def changed_project(self):
        html_path: Path = self.get_current_html_path()
        self.relation_id_to_relation_show_checkbox_dict.clear()
        if html_path.exists():

            self.load_html(html_path)
            global_vars.current_project.visualisation_uptodate.reset_relations_uptodate()
            self.update_slider_range()
            self.update_only_legend()

        else:
            # self.recreate_html(html_path)
            self.view.setHtml("")
            global_vars.current_project.visualisation_uptodate.set_clear_all(True)

        self.changed_project_bool = False

    def set_refresh_need_label_hidden_status(self,new_state:bool):
        self.refresh_needed_label.setHidden(new_state)

    def load_assets(self) -> List[OTLObject]:
        assets = RelationChangeDomain.get_visualisation_instances()
        return  assets

    def save_in_memory_changes_to_html(self):

        # TODO: fix the code to create the png form html (currently hangs the program)
        # hti = Html2Image(output_path=global_vars.current_project.get_project_local_path(),
        #                  size=(1920, 1080 ), )
        # hti.screenshot(
        #     html_file=str(self.get_current_html_path()),
        #     save_as="screenshot.png",
        #     size=(1920, 1080 )
        # )

        js_code = 'sendCurrentCombinedDataToPython();\n'
        self.view.page().runJavaScript(js_code)
        self.set_graph_saved_status(True)

    def set_graph_saved_status(self,saved:bool) -> None:
        if global_vars.current_project:
            global_vars.current_project.set_saved_graph_status(saved)

        if saved:
            self.graph_saved_status_label.setText(self._("Changes saved"))
            self.graph_saved_status_label.setStyleSheet('color:#11DD11;justify-content:right;')
        else:
            self.graph_saved_status_label.setText(self._("Changes are not saved"))
            self.graph_saved_status_label.setStyleSheet('color:#DD1111;justify-content:right;')

    def save_html(cls, file_path: Path,
                  new_node_data:str,
                  new_edge_data:str,
                  new_relationIdToSubEdges_data:str="",
                  new_relationIdToTotalSubEdgeCount_data:str="",
                  new_relationIdToJointNodes_data:str="",
                  new_SubEdgesToOriginalRelationId_data:str="",
                  new_edgeJointNodesIdToConnectionDataDict_data:str="",
                  new_collection_id_to_list_of_relation_ids_dict:str = ""
                  ) -> None:

        OTLLogger.logger.debug(
            f"Executing DataVisualisationScreen.save_html() for project {global_vars.current_project.eigen_referentie}",
            extra={"timing_ref": f"save_html_{global_vars.current_project.eigen_referentie}"})

        with open(file_path) as file:
            file_data = file.read()

        old_node_data = cls.extract_node_dataset(file_data)
        old_edge_data = cls.extract_edges_dataset(file_data)

        old_relationIdToSubEdges_data = cls.extract_support_data(
            variable_name="relationIdToSubEdges",
            file_data=file_data)
        old_relationIdToTotalSubEdgeCount_data = cls.extract_support_data(
            variable_name="relationIdToTotalSubEdgeCount",
            file_data=file_data)
        old_relationIdToJointNodes_data = cls.extract_support_data(
            variable_name="relationIdToJointNodes",
            file_data=file_data)
        old_SubEdgesToOriginalRelationId_data = cls.extract_support_data(
            variable_name="SubEdgesToOriginalRelationId",
            file_data=file_data)
        old_edgeJointNodesIdToConnectionDataDict_data = cls.extract_support_data(
            variable_name="edgeJointNodesIdToConnectionDataDict",
            file_data=file_data)

        old_collection_id_to_list_of_relation_ids_dict = cls.extract_support_dict_data(
            variable_name="collection_id_to_list_of_relation_ids",
            file_data=file_data)

        if old_node_data and old_edge_data:

            if old_node_data:
                file_data = cls.replace_node_data(old_node_data, new_node_data,file_data)
            if old_edge_data:
                file_data = cls.replace_edge_data(old_edge_data, new_edge_data,file_data)

            file_data = cls.replace_support_data(variable_name="relationIdToSubEdges",
                                                 old_data=old_relationIdToSubEdges_data,
                                                 new_data=new_relationIdToSubEdges_data,
                                                 file_data=file_data)
            file_data = cls.replace_support_data(variable_name="relationIdToTotalSubEdgeCount",
                                                 old_data=old_relationIdToTotalSubEdgeCount_data,
                                                 new_data=new_relationIdToTotalSubEdgeCount_data,
                                                 file_data=file_data)
            file_data = cls.replace_support_data(variable_name="relationIdToJointNodes",
                                                 old_data=old_relationIdToJointNodes_data,
                                                 new_data=new_relationIdToJointNodes_data,
                                                 file_data=file_data)
            file_data = cls.replace_support_data(variable_name="SubEdgesToOriginalRelationId",
                                                 old_data=old_SubEdgesToOriginalRelationId_data,
                                                 new_data=new_SubEdgesToOriginalRelationId_data,
                                                 file_data=file_data)
            file_data = cls.replace_support_data(variable_name="edgeJointNodesIdToConnectionDataDict",
                                                 old_data=old_edgeJointNodesIdToConnectionDataDict_data,
                                                 new_data=new_edgeJointNodesIdToConnectionDataDict_data,
                                                 file_data=file_data)
            file_data = cls.replace_support_dict_data(
                variable_name="collection_id_to_list_of_relation_ids",
                old_data=old_collection_id_to_list_of_relation_ids_dict,
                new_data=new_collection_id_to_list_of_relation_ids_dict,
                file_data=file_data)

            file_data = cls.disable_hierarchical_options(file_data)
            file_data = cls.disable_physics_option(file_data)
            file_data = cls.disable_stabilisation_loading_screen(file_data)

            with open(file_path,mode="w") as file:
                file.write(file_data)

            OTLLogger.logger.info(f"Saved html to: {file_path}")
            OTLLogger.logger.debug(
                f"Executing DataVisualisationScreen.save_html() for project {global_vars.current_project.eigen_referentie}",
                extra={"timing_ref": f"save_html_{global_vars.current_project.eigen_referentie}"})

        else:
            OTLLogger.logger.error(f"FAILED to save html: old node and edge data not found")

    @classmethod
    def replace_support_data(cls, variable_name, old_data, new_data, file_data):
        file_data = file_data.replace(
            f"var {variable_name} = new Map({old_data});",
            f"var {variable_name} = new Map({new_data});")
        return file_data

    @classmethod
    def replace_support_dict_data(cls, variable_name, old_data, new_data, file_data):
        file_data = file_data.replace(
            f"var {variable_name} = {old_data}",
            f"var {variable_name} = {new_data}")
        return file_data

    @classmethod
    def disable_physics_option(cls, file_data):
        new_physics_setting = ", \"physics\": {\"enabled\": false,\"stabilization\":{\"enabled\": false}, "
        if new_physics_setting not in file_data:
            file_data = file_data.replace(", \"physics\": {", new_physics_setting)
        return file_data

    @classmethod
    def disable_hierarchical_options(cls, file_data):
        file_data = file_data.replace("\"hierarchical\": {\"enabled\": true",
                                      "\"hierarchical\": {\"enabled\": false")
        return file_data

    @classmethod
    def disable_stabilisation_loading_screen(cls,file_data):
        new_code = (
            "var network = drawGraph();\n"
            "var loadingBar = document.getElementById('loadingBar');\n"
            "if (loadingBar) { loadingBar.style.display = 'none'; }"
        )
        if "var network = drawGraph();" in file_data:
            file_data = file_data.replace(
                "var network = drawGraph();",
                new_code
            )
        return file_data

    @classmethod
    def extract_node_dataset(cls, file_data):
        pattern = re.compile(
            r'nodes\s*=\s*new\s+vis\.DataSet\(([\s\S]*?)\);',
            re.DOTALL
        )
        m = pattern.search(file_data)
        if m:
            old_node_data = m.group(1)
        else:
            old_node_data = None
        return old_node_data

    @classmethod
    def replace_node_data(cls, old_data, new_data, file_data):
        # replace the whole node dataset statement to prevent wrong replacement
        return file_data.replace(f"nodes = new vis.DataSet({old_data})",
                                 f"nodes = new vis.DataSet({new_data})")

    @classmethod
    def replace_edge_data(cls, old_data, new_data, file_data):
        # replace the whole node dataset statement to prevent wrong replacement
        return file_data.replace(f"edges = new vis.DataSet({old_data})",
                                 f"edges = new vis.DataSet({new_data})")

    @classmethod
    def extract_edges_dataset(cls, file_data):
        pattern = re.compile(
            r'edges\s*=\s*new\s+vis\.DataSet\(([\s\S]*?)\);',
            re.DOTALL
        )
        m = pattern.search(file_data)
        if m:
            old_edges_data = m.group(1)
        else:
            old_edges_data = None
        return old_edges_data

    @classmethod
    def extract_support_data(cls, variable_name ,file_data):
        pattern = re.compile(
            fr'var\s*{variable_name}\s*=\s*new\s+Map\(([\s\S]*?)\);',
            re.DOTALL
        )
        m = pattern.search(file_data)
        if m:
            old_edges_data = m.group(1)
        else:
            old_edges_data = None
        return old_edges_data

    @classmethod
    def extract_support_dict_data(cls, variable_name, file_data):


        pattern = re.compile(
            fr'var\s*{variable_name}\s*=\s*([\s\S]*?);',
            re.DOTALL
        )
        m = pattern.search(file_data)
        if m:
            old_edges_data = m.group(1)
        else:
            old_edges_data = None
        return old_edges_data

    @classmethod
    def test_extract_support_dict_data(cls, regex:str, file_data):

        pattern = re.compile(
            regex,
            re.DOTALL
        )
        m = pattern.search(file_data)
        if m:
            old_edges_data = "{" + m.group(1) + "}"
        else:
            old_edges_data = None
        return old_edges_data


import asyncio

from PyQt6.QtWidgets import QMessageBox

from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception


class LoadingWindow(QMessageBox):



    def __init__(self, message:str = "Loading", title:str = "Loading", delayed_opening = False):
        super().__init__()
        self.message = message
        self.title = title
        self.opening = True
        self.delayed_opening = delayed_opening
        self.init_ui()

    def init_ui(self):
        self.setWindowTitle(self.title)
        self.setText(self.message)
        self.setStandardButtons(QMessageBox.NoButton)
        if self.delayed_opening:
            create_task_reraise_exception(self.delayed_open())
        else:
            self.open()

    def close(self):
        self.opening = False
        return super().close()

    async def delayed_open(self):
        # only open loading window after 1 second after initialisation
        # and only if the command to close it hasn't come yet by then
        await asyncio.sleep(1)
        if self.opening:
            self.open()






import logging

from PyQt6.QtGui import QColor
from PyQt6.QtWidgets import QApplication
from PyQt6.QtCore import Qt

from otlmow_gui.Domain import global_vars

from otlmow_gui.Domain.ProgramFileStructure import ProgramFileStructure


class Styling:
    #general styling
    light_stylesheet_filename = "custom_light.qss"
    dark_stylesheet_filename = "custom_dark.qss"

    #RelationScreen colors
    light_last_added_color = QColor("#d0ffcc")
    dark_last_added_color = QColor("#566b55")
    last_added_color = light_last_added_color

    #HeaderBar colors
    light_button_icon_color = QColor("#0E5A69")
    dark_button_icon_color = QColor("#FFFFFF")
    button_icon_color = light_button_icon_color

    @classmethod
    def applyStyling(cls,app:QApplication , meipass):
        """
        Apply the appropriate styling to the given application based on its color scheme. This method sets the application's stylesheet by loading the corresponding QSS file.

        :param app: The application instance to which the stylesheet will be applied.
        :type app: QApplication
        :param meipass: The path to the directory containing the executable, used to locate the stylesheet when running as a bundled application.
        :type meipass: str

        :return: None

        :raises FileNotFoundError: If the specified stylesheet file does not exist.

        :Examples:
            applyStyling(app, meipass)
        """

        style_library_name = "style"
        style_library_path = ProgramFileStructure.get_dynamic_library_path(style_library_name)

        if app.styleHints().colorScheme() == Qt.ColorScheme.Dark:
            style_filename = cls.dark_stylesheet_filename
            cls.last_added_color = cls.dark_last_added_color
            cls.button_icon_color = cls.dark_button_icon_color
        else:
            style_filename = cls.light_stylesheet_filename
            cls.last_added_color = cls.light_last_added_color
            cls.button_icon_color = cls.light_button_icon_color

        style_path = style_library_path / style_filename

        with open(style_path, 'r') as file:
            app.setStyleSheet(file.read())

        style_path_str = str(style_path.absolute())
        logging.debug(f"style sheet found in: {style_path_str}")

        #update a few hardcoded colors
        if global_vars.otl_wizard:
            global_vars.otl_wizard.main_window.update_color_scheme()





from pathlib import Path

from otlmow_modelbuilder.OSLOCollector import OSLOCollector

from otlmow_gui.Domain.database.SubsetDatabase import SubsetDatabase


class ModelBuilder:

    def __init__(self, path):
        self.collector = OSLOCollector(Path(path))
        self.subset_db = SubsetDatabase(Path(path))
        self.info_dict = dict(self.subset_db.get_general_info_project())
        if self.subset_db.is_valid_subset_database():
            self.collector.collect_all()
        self.close_subset_db()

    def get_all_classes(self):
        return self.collector.classes

    def filter_relations_and_abstract(self):
        relations = self.collector.relations
        relations_uri = {y.objectUri for y in relations}
        classes = self.get_all_classes()
        return [x for x in classes if x.abstract == 0 and x.objectUri not in relations_uri]

    def get_name_project(self):
        return self.info_dict.get('Naam')

    def get_operator_name(self):
        return self.info_dict.get('Operator')

    def get_otl_version(self):
        if 'BronVersie' in self.info_dict:
            # Subsets have a 'BronVersie' field to indicate the OTL version
            return self.info_dict.get('BronVersie')
        else:
            # Full OTL-model SQLite files have a 'Version' field to indicate the OTL version
            return self.info_dict.get('Version')

    def close_subset_db(self):
        self.subset_db.close_connection()



import os
import pathlib
import shutil
import sys

import PyInstaller.__main__

from otlmow_gui.Domain.logger.OTLLogger import OTLLogger

home_path = pathlib.Path.home()

OTLLogger.init()
OTLLogger.logger.debug("PyInstaller.__main__.run", extra={"timing_ref":"PyInstaller.__main__.run"})

root_path = pathlib.Path(__file__).parent.parent  # repo root
OTLLogger.logger.debug("root_path: " + str(root_path))
OTLLogger.logger.debug("sys.path: " + str(sys.path))

paths = None
for sys_path in sys.path:
    if 'site-packages' in sys_path:
        paths = sys_path


if not paths:
    raise FileNotFoundError("cannot find the site-packages folder")

OTLLogger.logger.debug("paths: " + paths)

sep = ';' if os.name == 'nt' else ':'

PyInstaller.__main__.run([
    str(root_path / 'otlmow_gui' / 'OTL Wizard 2.py'),
    '--distpath', str(root_path /  'LatestReleaseMulti'),
    '--contents-directory', 'data',
    '--paths', paths,
    '--paths', str(root_path),
    '--exclude-module','otlmow_model',
    '--collect-all', 'otlmow_converter',
    '--collect-all', 'otlmow_model',
    '--hidden-import', 'otlmow_model',
    '--collect-all', 'otlmow_template',
    '--collect-all', 'otlmow_modelbuilder',
    '--collect-all', 'otlmow_visuals',
    '--collect-all', 'pyvis',
    '--collect-all', 'numpy',
    '--collect-all', 'pandas',
    '--hidden-import', 'numpy',
    '--hidden-import', 'pandas',
    '--add-data', f"{str(root_path / 'otlmow_gui' / 'locale')}{sep}locale",
    '--add-data', f"{str(root_path / 'otlmow_gui' / 'style')}{sep}style",
    '--add-data', f"{str(root_path / 'otlmow_gui' / 'demo_projects')}{sep}demo_projects",
    '--add-data', f"{str(root_path / 'otlmow_gui' / 'img')}{sep}img",
    '--add-data', f"{str(root_path / 'pyproject.toml')}:.",
    '--add-data', f"{str(root_path / 'otlmow_gui' / 'javascripts_visualisation')}{sep}javascripts_visualisation",
    '--icon',f"{str(root_path / 'otlmow_gui' / 'img')}/wizard.png",
    '--splash',f"{str(root_path / 'otlmow_gui' / 'img')}/Logo-OTL_Wizard_2_no_purple_edge_V3.png",
    '--noconfirm',
    # '--onefile',  # All files packed together in one executable file
    '--noconsole', # no cmd/powershell window with debug output
    '--clean'
])

OTLLogger.logger.debug("PyInstaller.__main__.run", extra={"timing_ref":"PyInstaller.__main__.run"})

# rename dist folder
original_folder = root_path / 'LatestReleaseMulti' / 'OTL Wizard 2'
target_folder = root_path / 'LatestReleaseMulti' / 'OTL_Wizard_2'
if original_folder.exists():
    if target_folder.exists():
        shutil.rmtree(target_folder)
    shutil.move(str(original_folder), str(target_folder))
else:
    raise FileNotFoundError(f"Expected folder not found: {original_folder}")

# final executable path
exe_path = target_folder / 'OTL Wizard 2.exe'
if not exe_path.exists():
    raise FileNotFoundError(f"Executable missing: {exe_path}")
print(f"Executable at: {exe_path}")

# TODO: make sure custom.qss and demo projects are copied next to .exe > verify it is loaded same as locale


from abc import abstractmethod
from copy import deepcopy
from pathlib import Path
from typing import Optional

from PyQt6.QtWidgets import QFileDialog

from otlmow_gui.Domain import global_vars


class AbstractFilePickerDialog(QFileDialog):


    def __init__(self, language_settings, title: Optional[str], name_filter: str = None,
                 initial_file_path: Optional[str] = None):
        super().__init__()
        self._ = language_settings
        self.name_filter = name_filter
        self.previous_file_dialog_state = None
        self.previous_exported_file_name = None

        self.supported_export_formats: dict = deepcopy(global_vars.supported_file_formats)

        self.setWindowTitle(title)
        if initial_file_path:
            self.setDirectory(initial_file_path)
        else:
            self.setDirectory(str(Path.home() / "Documents"))
        self.setNameFilter(self.getCustomFilters())

    def getCustomFilters(self) -> str:
        if self.name_filter:
            return self.name_filter

        supported_extensions = [f'*.{value}' for value in
                                self.supported_export_formats.values()]
        supported_extensions_string = ' '.join(supported_extensions)
        filters = [self._('Alle ondersteunde bestanden ({supported_extensions_string})').format(
            supported_extensions_string=supported_extensions_string)]

        filters.extend(f"{k} files (*.{v})" for k, v in self.supported_export_formats.items())

        return ";;".join(filters)

    def summon(self, chosen_file_format: Optional[str]= "", directory: Optional[str|Path]=None,
               supported_export_formats:dict[str,str]=None,project_name:Optional[str] = None):

        if directory:
            if isinstance(directory,str):
                directory = Path(directory)
            if directory.exists():
                self.setDirectory(str(directory))
            elif self.previous_file_dialog_state:
                self.setDirectory(self.previous_file_dialog_state)

        elif self.previous_file_dialog_state:
            self.setDirectory(self.previous_file_dialog_state)

        if chosen_file_format:
            if not supported_export_formats:
                supported_export_formats = self.supported_export_formats

            if chosen_file_format in supported_export_formats:
                file_suffix = supported_export_formats[chosen_file_format]
                filter_filepicker = f"{chosen_file_format} files (*.{file_suffix})"
                self.setNameFilter(filter_filepicker)

        self.set_filename_suggestions(project_name)

        res = self.execute()

        self.previous_file_dialog_state = self.directory().absolutePath()
        self.store_save_filename(res)

        #make sure that loaded files actually have the format that is chosen
        if chosen_file_format and (chosen_file_format in supported_export_formats) and res:
            file_suffix = supported_export_formats[chosen_file_format]

            for i in range(len(res)):
                res[i] = res[i].with_suffix("")
                if res[i].suffix != f".{file_suffix}":
                    res[i]= res[i].with_suffix(f".{file_suffix}")

        return res

    def set_filename_suggestions(self, project_name):
        pass

    @abstractmethod
    def execute(self) -> list[Path]:
       return []

    def store_save_filename(self,res:list[str]):
        pass

    def get_filename_suggestion(self, project_name):
        pass


from PyQt6.QtWidgets import QVBoxLayout

from otlmow_gui.GUI.screens.ExportData_elements.AbstractExportDataSubScreen import AbstractExportDataSubScreen



class ExportAllDataSubScreen(AbstractExportDataSubScreen):
    def __init__(self, language_settings=None):
        super().__init__(language_settings=language_settings)

    def init_ui(self) -> None:
        """
        Initializes the user interface for the export data screen. This method sets up the layout
        by adding spacing, a menu, and configuring the container's margins and stretch properties
        to ensure proper alignment and appearance.

        :return: None
        """



        sub_window_layout = QVBoxLayout()
        sub_window_layout.addWidget(self.create_filetype_combobox())
        sub_window_layout.addWidget(self.create_relations_split_optionality())
        sub_window_layout.addSpacing(10)
        sub_window_layout.addWidget(self.create_button_box())
        sub_window_layout.addSpacing(10)
        sub_window_layout.addStretch()
        self.setLayout(sub_window_layout)



import asyncio
import base64
import pathlib

from PyQt6.QtCore import QUrl
from folium import folium, JsCode

from otlmow_gui.Domain.logger.OTLLogger import OTLLogger

from shapely.wkt import loads
from pyproj import Transformer


class MapHelper:
    added_layer_asset_id_list = []
    normal_color = "#FF7F00"
    highlight_color = "#004d5c"

    @classmethod
    async def create_html_map(cls,id_to_object_with_text_and_data_dict:dict,IMG_DIR,HTML_DIR,prev_selected_asset_id=None):
        m = cls.create_folium_map()

        OTLLogger.logger.debug(pathlib.Path.home().drive)
        OTLLogger.logger.debug(IMG_DIR)
        OTLLogger.logger.debug(str(
            (IMG_DIR/ "bol_orange.png")))

        img_qurl = QUrl(pathlib.Path.home().drive + str(
            (IMG_DIR / "bol_orange.png").absolute()).replace("\\", "/"))
        img_path = img_qurl.path()

        with open(img_path, 'rb') as img_file:
            b64_data = base64.b64encode(img_file.read()).decode('utf-8')

        # Create the data URL
        data_url = f"data:image/png;base64,{b64_data}"

        click_js = ("function onMapClick(e) {\n"
                    f"var icon_path = '{data_url}';\n ")
        click_js += """
                               
                                var map_id = e.originalEvent.srcElement.id;
                                var lat = e.latlng.lat;
                                var lng = e.latlng.lng;
             
                                //drawPoint(lat,lng,map_id,"","");
                                

                                //console.log(e);-

                                //drawing polygons
                                var lat1 = lat+1;
                                var lng1 = lng+1;

                                //// var latlngs = [[lat, lng],[lat1, lng],[lat1, lng1],[lat, lng1]];// sqaure
                                //// var latlngs = [[lat, lng],[lat1, lng]];// line
                                var latlngs = [[lat, lng],[lat1, lng],[lat1, lng1]];// corner
                                ////var latlngs = [[lat, lng]];// dot
                                //var map = e.originalEvent.srcElement;
                                
                                ////var polygon = L.polygon(latlngs, {color: "red"}).addTo(eval(map_id));
                                //// zoom the map to the polygon
                                ////eval(e.originalEvent.srcElement.id).fitBounds(polygon.getBounds());
                                
                                //drawLines(latlngs, map_id);

                                // Send coordinates to Python
                                //activateHighlightLayer( 'e64e7fcb-d429-4e3d-8651-706297f14ca4-b25kZXJkZWVsI1ZvZXJ0dWlnbGFudGFhcm4');
                                
                            //}
                             activateHighlightLayer( 'a303e80b-9863-4b47-b0c0-dadb8fc9b651-b25kZXJkZWVsI1ZvZXJ0dWlnbGFudGFhcm4');
                            if('a303e80b-9863-4b47-b0c0-dadb8fc9b651-b25kZXJkZWVsI1ZvZXJ0dWlnbGFudGFhcm4' && ('a303e80b-9863-4b47-b0c0-dadb8fc9b651-b25kZXJkZWVsI1ZvZXJ0dWlnbGFudGFhcm4' in idToLayerDict))
                            {
                              eval(map_id).fitBounds(idToLayerDict['a303e80b-9863-4b47-b0c0-dadb8fc9b651-b25kZXJkZWVsI1ZvZXJ0dWlnbGFudGFhcm4'].getBounds());
                              }
                        """

        click_js += "}"
        m.add_js_link(name="QWebChannel_script", url="qrc:///qtwebchannel/qwebchannel.js")
        m.on(click=JsCode(click_js))

        js_webchannel_script =  (f"var mapEl = document.getElementById('{m.get_name()}');\n"
                                f"var icon_path = '{data_url}';\n "
                                 f"var normal_color = '{cls.normal_color}';\n"
                                 f"var highlight_color = '{cls.highlight_color}';\n")
        js_webchannel_script += """
        
        //global state vars
        var previousSelectedId = null;
        var idToLayerDict = {}
        var featureGroup = L.featureGroup()
        
        //turn cursor/mousepointer into crosshair
        mapEl.style.cursor = "crosshair";
        
        // add webchannel to javascript to communicate with python
        try{
            var channel = new QWebChannel(qt.webChannelTransport, function(channel) {
                window.pywebchannel = channel.objects.webBridge;
            });
        } catch (error) {
          console.error(error);
          alert("MapScreen: QWebChannel creation error");
        }
        
        
        function drawLines(latlngs, map_id, text, id)
        {
            var line = L.polyline(latlngs, {color:normal_color, weight:5}).addTo(eval(map_id)).bindPopup(text,{autoPan:false});
            idToLayerDict[id] = line
            line.on('click', function (e)
            {
                //highlight on click
                var layer = e.target;
                activateHighlightLayer(id);
                sendSelectedIdToPython(id);
            
            });
            featureGroup.addLayer(line);
        }
        
        function drawPolygons(latlngs, map_id, text, id)
        {
            var line = L.polygon(latlngs, {color:normal_color, weight:5}).addTo(eval(map_id));
            line.bindPopup(text,{offset:L.point(0,0),autoPan:false})
            idToLayerDict[id] = line;
            line.on('click', function (e)
            {
                //highlight on click
                var layer = e.target;
                activateHighlightLayer(id);
                sendSelectedIdToPython(id);
             
            });
            featureGroup.addLayer(line);
        }
        
        function drawPoint(lat, lng, map_id, text, id)
        {
            var bolIcon = L.icon(
            {
                    iconUrl: icon_path,
                    iconSize:     [15, 15], // size of the icon
                    iconAnchor:   [7.5, 7.5], // point of the icon which will correspond to marker's location
                    popupAnchor:  [0.5, 3] // point from which the popup should open relative to the iconAnchor
            });
            // Add marker dynamically
            var marker = L.marker([lat, lng], {icon: bolIcon}).addTo(eval(map_id))
            .bindPopup(text,{autoPan:false})
            //.openPopup();
            idToLayerDict[id] = marker
           
            // sendCoordinatesToPython(lat,lng);
            marker.on('click', function (e)
            {
                //highlight on click
                var layer = e.target;
                activateHighlightLayer(id);
                sendSelectedIdToPython(id);
           
            });
            featureGroup.addLayer(marker);
        }
        
        function activateHighlightLayer(id)
        {
            if(id in idToLayerDict)
            {
                var layer = idToLayerDict[id];
                var htmlElement = undefined
                if (layer._icon == undefined)
                {
                    //if the visual is line
                    htmlElement = layer._path
                    if(layer.options.color == normal_color)
                    {
                        disablePreviousHighlightLayer();
                        layer.setStyle({color: highlight_color, weight:8})
                        layer.options.color = highlight_color;
                        layer.options.weight = 8;
                        layer.redraw()
                        previousSelectedId = id;
                        layer.openPopup();
                    }
                }
                else
                {
                    //if the visual is marker
                    htmlElement = layer._icon
                
                    if(!L.DomUtil.hasClass(htmlElement, 'dash-border'))
                    {
                        disablePreviousHighlightLayer();
                        L.DomUtil.addClass(htmlElement,'dash-border');
                        previousSelectedId = id
                        layer.openPopup();
                    }
                }
            }
            else
            {
                disablePreviousHighlightLayer();
                previousSelectedId = id
            }    
        }
        
        function sendCoordinatesToPython(lat,lng)
        {
            if (window.pywebchannel) 
            {
                window.pywebchannel.receive_coordinates(JSON.stringify({lat: lat, lng: lng}));
            } 
            else 
            {
                console.log("QWebChannel is not initialized yet.");
                alert("MapScreen: QWebChannel is not initialized");
            }
        }
        
        function sendSelectedIdToPython(id)
        {
            if (window.pywebchannel) 
            {
                window.pywebchannel.receive_selection_id(JSON.stringify({id: id}));
            } 
            else 
            {
                console.log("QWebChannel is not initialized yet.");
            }
        }
        
        function disablePreviousHighlightLayer()
        {
            if(previousSelectedId && (previousSelectedId in idToLayerDict))
            {
                var prevLayer = idToLayerDict[previousSelectedId]
                var prevLayerhtmlElement = undefined
                if (prevLayer._icon == undefined)
                {
                    //if the previous visual is line
                    prevLayer.setStyle({color: normal_color, weight:5})
                    prevLayer.options.color = normal_color
                    prevLayer.options.weight = 5
                    prevLayer.redraw()
                }
                else
                {
                    //if the previous visual is marker
                    prevLayerhtmlElement = prevLayer._icon
                    L.DomUtil.removeClass(prevLayerhtmlElement,'dash-border');
                }
                    
            }
        }
        
        function goToLayer(id, map_id)
        {
            if(id && (id in idToLayerDict))
            {
                var layer = idToLayerDict[id];
                if (layer._icon == undefined)
                {
                    eval(map_id).fitBounds(layer.getBounds(),{maxZoom:20.5});
                }
                else
                {
                    //this means this is a marker (point geometry)
                    eval(map_id).panTo(layer.getLatLng())
                }
            }
        }
        
        
        
            """

        m.get_root().script.add_child(folium.Element(js_webchannel_script))
        marker_count = len(id_to_object_with_text_and_data_dict)
        OTLLogger.logger.debug(f"Adding markers to map ({marker_count} markers",extra={"timing_ref":"adding_marker_to_map"})
        init_script = 'document.addEventListener("DOMContentLoaded", (event) => {\n'
        for id, otl_object_with_text_and_data in id_to_object_with_text_and_data_dict.items():
            await asyncio.sleep(0)
            otl_object = otl_object_with_text_and_data[0]

            text_and_data = otl_object_with_text_and_data[1]

            if hasattr(otl_object, "geometry") and otl_object.geometry:
                if "POINT" in otl_object.geometry:

                    transformed_geometry = MapHelper.convert_wkt_to_wgs84(otl_object.geometry)
                    coord_list = MapHelper.extract_first_level(transformed_geometry)
                    # MapHelper.addProjectedMarker()
                    popup_text = text_and_data["text"].screen_name + "<br>" + text_and_data[
                        "text"].typeURI #+ "<br>" + otl_object.geometry
                    for coord in coord_list:
                        init_script += MapHelper.add_projected_marker(coord, m.get_name(),popup_text, id)

                if "LINESTRING" in otl_object.geometry:
                    transformed_geometry = MapHelper.convert_wkt_to_wgs84(otl_object.geometry)
                    coord_list = MapHelper.extract_first_level(transformed_geometry)

                    popup_text = text_and_data["text"].screen_name + "<br>" + text_and_data[
                        "text"].typeURI #+ "<br>" + otl_object.geometry
                    for pair in coord_list:
                        init_script += MapHelper.add_projected_line(pair.split(","), m.get_name(),popup_text, id)

                if "POLYGON" in otl_object.geometry:
                    transformed_geometry = MapHelper.convert_wkt_to_wgs84(otl_object.geometry)
                    polygon_list = MapHelper.extract_first_level(transformed_geometry)
                    coord_list = []
                    for polygon in polygon_list:
                        second_level_list = MapHelper.extract_first_level(polygon)
                        for coordinates in second_level_list:
                            coord_list.append(coordinates.split(","))

                    popup_text = text_and_data["text"].screen_name + "<br>" + text_and_data[
                        "text"].typeURI
                    init_script += MapHelper.add_projected_polygon(coord_list, m.get_name(),
                                                                popup_text, id)

        OTLLogger.logger.debug(f"Added markers to map ({marker_count} markers",
                               extra={"timing_ref": "adding_marker_to_map"})
        init_script += cls.get_zoom_to_assets_js_code(map_id=m.get_name(), prev_selected_asset_id=prev_selected_asset_id)
        init_script += "});"

        # make highlight style when clicked
        m.get_root().script.add_child(folium.Element(init_script))
        dash_style = ("""
        <style>
        .dash-border {"""
            f"border: 2px dashed {cls.highlight_color};"
            f"background-color: {cls.highlight_color};"
        """}
        
        
        </style>
        """)
        m.get_root().header.add_child(folium.Element(dash_style))

        map_path = cls.get_map_html_save_path(HTML_DIR, IMG_DIR)
        m.save(map_path)
        map_id = m.get_name()
        map = m
        return map_path, map , map_id

    @classmethod
    def get_map_html_save_path(cls, HTML_DIR: pathlib.Path, IMG_DIR: pathlib.Path) -> str:
        html_dir = IMG_DIR.parent / 'html'
        if not html_dir.exists():
            html_dir = HTML_DIR

        return str((html_dir / "folium_map.html").absolute())

    @classmethod
    def create_folium_map(cls):
        coordinate = (51.16872907594677, 4.41375968966803)# antwerpen
        cls.added_layer_asset_id_list.clear()
        # satelite image layer from google maps instead of open street road map
        tile = folium.TileLayer(
            tiles='https://geo.api.vlaanderen.be/OMW/wmts?SERVICE=WMTS&VERSION=1.0.0&REQUEST=GetTile&LAYER=omwrgb24vl&STYLE=&FORMAT=image/png&TILEMATRIXSET=GoogleMapsVL&TILEMATRIX={z}&TILEROW={y}&TILECOL={x}',
            attr='GRB',
            name='GRB Satellite',
            maxZoom=23,
            maxNativeZoom=21,
            overlay=False,
            control=True
        )
        return folium.Map(
            zoom_start=13,
            location=coordinate,
            tiles=tile
        )

    @classmethod
    def add_marker(cls, lat, lon, map_id, web_view):
        """Adds a marker dynamically without reloading the map."""
        # js_code = f'L.marker([{lat}, {lon}]).addTo({self.map_id}).bindPopup("Marker at {lat}, {lon}").openPopup();'
        """add a polygon sqaure to the map and go to it"""
        insert = '{color: "red"}'
        # js_code =f'var polyline = L.polygon([[{lat}, {lon}],[{lat+1}, {lon}],[{lat+1}, {lon+1}],[{lat}, {lon+1}]],{insert} ).addTo(eval({self.map_id}));\neval({self.map_id}).fitBounds(polyline.getBounds());'
        js_code = f"drawPoint({lat},{lon},{map_id},"");"
        web_view.page().runJavaScript(js_code)

    @classmethod
    def add_projected_marker(cls, coord:str, map_id:str, tooltip_text:str,id:str):
        """Adds a marker dynamically without reloading the map."""

        js_code = (f"var split = '{coord}'.split(' ');\n"
            "var point = L.CRS.EPSG3857.unproject(L.point([parseFloat(split[0]), parseFloat(split[1])] ));\n"
                   f"drawPoint(point.lat,point.lng,'{map_id}','{tooltip_text}','{id}');\n")

        cls.added_layer_asset_id_list.append(id)

        return js_code

    @classmethod
    def add_projected_line(cls, coord_pair: str, map_id: str, tooltip_text: str, id: str):
        """Adds a marker dynamically without reloading the map."""

        js_code = (f"var split_pair = {coord_pair}\n"
                   "latlngs = [];\n"
                   "for (const  coord of split_pair) \n"
                   "{\n"
                   "    //console.log(coord);\n"
                   "    var split = coord.trim().split(' ');\n"
                   "    var point = L.CRS.EPSG3857.unproject(L.point([parseFloat(split[0].trim()), parseFloat(split[1].trim())] ));\n"
                   "    //console.log(point);\n"
                   "    latlngs.push([point.lat,point.lng]);\n"
                   "}\n"
                    f"drawLines(latlngs,'{map_id}','{tooltip_text}','{id}');\n")

        cls.added_layer_asset_id_list.append(id)

        return js_code

    @classmethod
    def add_projected_polygon(cls, coord_pair: list, map_id: str, tooltip_text: str, id: str):
        """Adds a marker dynamically without reloading the map."""

        js_code = (f"var polygons = {coord_pair}\n"
                   "for (const  split_pair of polygons) \n"
                   "{\n"
                       "    latlngs = [];\n"
                       "    for (const  coord of split_pair) \n"
                       "    {\n"
                       "        //console.log(coord);\n"
                       "        var split = coord.trim().split(' ');\n"
                       "        var point = L.CRS.EPSG3857.unproject(L.point([parseFloat(split[0].trim()), parseFloat(split[1].trim())] ));\n"
                       "        //console.log(point);\n"
                       "        latlngs.push([point.lat,point.lng]);\n"
                       "    }\n"
                       f"   drawPolygons(latlngs,'{map_id}','{tooltip_text}','{id}');\n"
                    "}\n")
        cls.added_layer_asset_id_list.append(id)

        return js_code

    @classmethod
    def activate_highlight_layer_by_id(cls, asset_id, web_view,map_id):

        js_code = ( f"activateHighlightLayer( '{asset_id}');\n"
                    f"goToLayer('{asset_id}', '{map_id}');\n")

        web_view.page().runJavaScript(js_code)

    @classmethod
    def extract_first_level(cls,s):
        results = []
        level = 0
        start = None
        for i, char in enumerate(s):
            if char == '(':
                level += 1
                if level == 1:  # starting a new top-level group
                    start = i + 1  # content starts after the '('
            elif char == ')':
                if level == 1 and start is not None:
                    results.append(s[start:i])
                level -= 1
        return results

    @classmethod
    def convert_wkt_to_wgs84(cls,wkt_string):
        # Load the geometry from the WKT string
        geometry = loads(wkt_string)

        # Define the source CRS (EPSG:31370)
        source_crs = 'EPSG:31370'

        # Define the target CRS (WGS84)
        target_crs = 'EPSG:3857'

        transformer = Transformer.from_crs(crs_from=source_crs, crs_to=target_crs)
        # Transform the coordinates
        if geometry.geom_type == 'Point':
            x, y = geometry.x, geometry.y


            transformer.transform(12, 12)
            lon, lat = transformer.transform( x, y)
            return f'POINT ({lon} {lat})'

        elif geometry.geom_type in ['LineString', 'MultiPoint']:
            transformed_coords = [
                transformer.transform( x, y) for x, y, z in geometry.coords
            ]
            return f'LINESTRING ({", ".join(f"{lon} {lat}" for lon, lat in transformed_coords)})'

        elif geometry.geom_type in ['Polygon', 'MultiLineString']:
            transformed_coords = [
                transformer.transform( x, y) for x, y,z in geometry.exterior.coords
            ]
            return f'POLYGON (({" , ".join(f"{lon} {lat}" for lon, lat in transformed_coords)}))'

        elif geometry.geom_type == 'MultiPolygon':
            transformed_polygons = []
            for polygon in geometry.geoms:
                transformed_coords = [
                    transformer.transform( x, y) for x, y in polygon.exterior.coords
                ]
                transformed_polygons.append(
                    f'POLYGON (({" , ".join(f"{lon} {lat}" for lon, lat in transformed_coords)}))')
            return f'MULTIPOLYGON ({", ".join(transformed_polygons)})'

        else:
            raise ValueError(f"Unsupported geometry type: {geometry.geom_type}")

    @classmethod
    def zoom_to_assets(cls,web_view,map_id, prev_selected_asset_id=None):
        js_code = cls.get_zoom_to_assets_js_code(map_id=map_id,prev_selected_asset_id=prev_selected_asset_id)
        web_view.page().runJavaScript(js_code)

    @classmethod
    def get_zoom_to_assets_js_code(cls,map_id, prev_selected_asset_id):
        js_code = f"var previousSelectedId = '{prev_selected_asset_id}'"
        js_code += ("""
                  if(previousSelectedId && previousSelectedId in idToLayerDict)
                  {
                      activateHighlightLayer(previousSelectedId)
                  """)
        js_code += f"eval('{map_id}').fitBounds(idToLayerDict[previousSelectedId].getBounds());"
        js_code += ("""
                  }
                  else if(featureGroup)
                  {
                 """)
        js_code += f"eval('{map_id}').fitBounds(featureGroup.getBounds());"
        js_code +="}"

        return js_code




from copy import deepcopy
from pathlib import Path

from otlmow_model.OtlmowModel.Classes.ImplementatieElement.RelatieObject import RelatieObject
from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.util.Helpers import Helpers
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.step_domain.RelationChangeDomain import RelationChangeDomain
from otlmow_gui.GUI.dialog_windows.NotificationWindow import NotificationWindow
from otlmow_gui.GUI.translation.GlobalTranslate import GlobalTranslate


class ExportDataDomain:
    """
    Handles the generation and management of export files for assets and relations.
    This class provides methods to create, split, and organize data for export, ensuring that the
    output meets specified requirements.

    Attributes:
        None

    Methods:
        generate_files(end_file: Path, separate_per_class_csv_option: bool = False, separate_relations_option: bool = False) -> None: Generates output files based on the current assets and relations in memory.
        split_relations_and_objects(objects_in_memory: list) -> tuple: Splits a list of objects into two categories: assets and relations.
        create_relation_and_asset_path(end_file: Path | str) -> tuple: Creates file paths for relations and assets based on the provided end file path.
    """


    @classmethod
    async def generate_files(cls, end_file: Path, separate_per_class_csv_option : bool =False,
                       separate_relations_option:bool =False, **kwargs) -> None:
        """
        Generates output files based on the current assets and relations in memory.
        This class method allows for the option to separate relations and assets into
        different files, and can also split the output by class type if specified.

        Args:
            end_file (Path): The path where the output file(s) will be saved.
            separate_per_class_csv_option (bool, optional): If True, splits the output into separate CSV files per class type. Defaults to False.
            separate_relations_option (bool, optional): If True, generates separate files for relations and assets. Defaults to False.

        Returns:
            None

        Raises:
            OSError: If there is an issue writing to the specified output file(s).
        """
        try:
            assets_in_memory = sorted(RelationChangeDomain.get_internal_objects(), key=lambda relation1: relation1.typeURI)
            relations_in_memory = sorted(RelationChangeDomain.get_persistent_relations(), key=lambda relation1: relation1.typeURI)
            await cls.export_to_files(assets_in_memory, relations_in_memory, end_file,
                                separate_per_class_csv_option, separate_relations_option, **kwargs)
            cls.get_screen().open_folder_of_created_export_files(end_file)
        except ValueError as e:
            if e.args == ('There are no asset data to export to Excel',):
                notification = NotificationWindow(title="Geen data",
                                                  message="Er is geen data om te exporteren")
                notification.exec()
            else:
                raise e

    @classmethod
    async def export_to_files(cls, assets, relations, end_file, separate_per_class_csv_option,
                        separate_relations_option, **kwargs):
        try:
            if separate_relations_option:
                relations_path, assets_path = cls.create_relation_and_asset_path(end_file)
                if relations:
                    await Helpers.start_async_converter_from_object_to_file(file_path=relations_path,
                                                                      sequence_of_objects=relations,
                                                                      split_per_type=separate_per_class_csv_option,
                                                                      abbreviate_excel_sheettitles=True, **kwargs)
                else:
                    OTLLogger.logger.info(
                        f"No Relations in memory for project {global_vars.current_project.eigen_referentie}")
                if assets:
                    await Helpers.start_async_converter_from_object_to_file(file_path=assets_path,
                                                                      sequence_of_objects=assets,
                                                                      split_per_type=separate_per_class_csv_option,
                                                                      abbreviate_excel_sheettitles=True, **kwargs)
                else:
                    OTLLogger.logger.info(
                        f"No Assets in memory for project {global_vars.current_project.eigen_referentie}")

            else:
                objects_in_memory = deepcopy(assets)
                objects_in_memory.extend(relations)
                await Helpers.start_async_converter_from_object_to_file(file_path=Path(end_file),
                                                                  sequence_of_objects=objects_in_memory,
                                                                  split_per_type=separate_per_class_csv_option,
                                                                  abbreviate_excel_sheettitles=True, **kwargs)
        except PermissionError as e:
            notification = NotificationWindow(title= GlobalTranslate._("Toestemming geweigerd"),
                                              message= GlobalTranslate._("Geen toestemming om te exporteren naar \n{filename}\nIs het bestand nog open?").format(filename=e.filename))
            notification.exec()
            return
    @classmethod
    def split_relations_and_objects(cls,objects_in_memory):
        """
        Splits a list of objects into two categories: assets and relations. This method iterates
        through the provided objects, classifying them based on their attributes, and returns
        two separate lists for further processing.

        :param objects_in_memory: A list of objects to be classified into assets and relations.
        :type objects_in_memory: list

        :return: A tuple containing two lists: the first list of assets and the second list of relations.
        :rtype: tuple[list, list]
        """

        OTLLogger.logger.debug("started splitting relations and objects for export")
        assets_in_memory = []
        relations_in_memory = []

        for obj in objects_in_memory:
            if hasattr(obj, 'bronAssetId') or isinstance(obj, RelatieObject):
                relations_in_memory.append(obj)
            else:
                assets_in_memory.append(obj)
        return assets_in_memory, relations_in_memory

    @classmethod
    def create_relation_and_asset_path(cls, end_file):
        """
        Creates file paths for relations and assets based on the provided end file path.
        This class method constructs new file names by appending appropriate suffixes to the original file name and returns the paths for both relations and assets.

        Args:
            end_file (Path | str): The path of the original file used to derive the new file paths.

        Returns:
            tuple: A tuple containing the paths for the relations file and the assets file.
        """

        OTLLogger.logger.debug("started creating relation and class path for export")
        parent_directory = Path(end_file).parent
        file_stem = Path(end_file).stem
        file_suffix = Path(end_file).suffixes[0]

        relations_file_name = f"{file_stem}_relations{file_suffix}"
        assets_file_name = f"{file_stem}_assets{file_suffix}"

        relations_path = Path(parent_directory) / relations_file_name
        assets_path = Path(parent_directory) / assets_file_name
        return relations_path, assets_path

    @classmethod
    def update_frontend(cls):
        pass

    @classmethod
    def get_screen(cls):
        return global_vars.otl_wizard.main_window.step4_export


import datetime
import logging
import os
import traceback
import warnings

from pathlib import Path

from otlmow_gui.Domain.ProgramFileStructure import ProgramFileStructure


class OTLLogger(logging.Logger):

    logger = None
    ref_key_to_time_dict = {}
    loading_window = None


    def __init__(self, name, level=0):
        super().__init__(name, level)

    @classmethod
    def init(cls):
        cls.logger = cls(__name__)

        logging_file = cls.create_logging_file()
        cls.remove_old_logging_files()

        file_handler = logging.FileHandler(logging_file) # logger will write to file
        stderr_handler = logging.StreamHandler() # logger will write to stderr
        #
        cls.logger.addHandler(stderr_handler)
        cls.logger.addHandler(file_handler)

        logging.getLogger().addHandler(stderr_handler)
        logging.getLogger().addHandler(file_handler) # loggers from other libraries will write to file

        # cls.logger = logging.getLogger()
        cls.logger.setLevel(logging.DEBUG)

        # Warning are written to console and to file
        # Function to redirect warnings to the logger
        def log_runtime_warnings(message, category, filename, lineno, file=None, line=None):
            cls.logger.warning(f"{category.__name__}: {message} (File: {filename}, Line: {lineno})")
        # Redirect warnings to logger
        warnings.showwarning = log_runtime_warnings

        for handler in cls.logger.handlers:
            handler.setFormatter(logging.Formatter(
                fmt='%(asctime)s %(message)s',
                                   datefmt='%Y-%m-%d %H:%M:%S'))

    @classmethod
    def create_logging_file(cls) -> Path:

        work_dir_path = ProgramFileStructure.get_otl_wizard_work_dir() / 'logs'
        if not work_dir_path.exists():
            work_dir_path.mkdir()
        timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        logging_filepath = work_dir_path / f'OTL_wizard_2_{timestamp}.log'
        if not logging_filepath.exists():
            open(Path(logging_filepath), 'w').close()
        return logging_filepath

    @classmethod
    def remove_old_logging_files(cls) -> None:

        work_dir_path = ProgramFileStructure.get_otl_wizard_work_dir() / 'logs'
        for file in os.listdir(work_dir_path):
            file_name = Path(file).stem
            split_file = file_name.split('_')

            date_str = str(split_file[-1])
            # cls.logger.debug(f"split file date {date_str}")
            datetime_obj = datetime.datetime.strptime(split_file[-1], "%Y%m%d%H%M%S")
            if datetime_obj < datetime.datetime.now() - datetime.timedelta(days=7):
                os.remove(work_dir_path / file)

    def _alter_msg(self,msg,extra):

        ref_key = None
        if extra and "timing_ref" in extra :
            ref_key = extra["timing_ref"]

        if not  ref_key:
            # return "{status:5s}({time:07.3f}s) {msg}".format(status="", time=0,msg=msg)
            return msg

        current_time = datetime.datetime.now()
        if ref_key and ref_key in OTLLogger.ref_key_to_time_dict:
            ref_time = OTLLogger.ref_key_to_time_dict.pop(ref_key)
            time_dif:datetime.timedelta = current_time - ref_time
            time_dif_seconds = time_dif.seconds + time_dif.microseconds/1000000
            state = "END"

            return "{msg} {status:5s} {ref} ({time:07.3f}s) ".format(status=state, time=time_dif_seconds,
                                                              ref=ref_key, msg=msg)
        else:
            state = "START"
            OTLLogger.ref_key_to_time_dict[ref_key] = current_time
            # OTLLogger.attempt_show_loading_screen(ref_key)
            time_dif_seconds = 0
            return "{msg} {status:5s} {ref}".format(status=state,
                                                              msg=msg,ref=ref_key)




    def debug(self, msg, *args, exc_info=None, stack_info=False, stacklevel=1, extra=None):

        msg = self._alter_msg(msg,extra)
        stack =traceback.extract_stack(f=None, limit=2)
        filename = Path(stack[0].filename).name
        lineno = stack[0].lineno
        levelname = "DEBUG"
        super().debug(f"ln {lineno:4d}:{filename:25s}  {levelname:8s}  {msg}" , *args, exc_info=exc_info, stack_info=stack_info, stacklevel=stacklevel,
                      extra=extra)

    def info(self, msg, *args, exc_info=None, stack_info=False, stacklevel=1, extra=None):
        msg = self._alter_msg(msg, extra)
        stack = traceback.extract_stack(f=None, limit=2)
        filename = Path(stack[0].filename).name
        lineno = stack[0].lineno
        levelname = "INFO"
        super().info(f"ln {lineno:4d}:{filename:25s}  {levelname:8s}  {msg}", *args, exc_info=exc_info, stack_info=stack_info, stacklevel=stacklevel,
                     extra=extra)

    def warning(self, msg, *args, exc_info=None, stack_info=False, stacklevel=1, extra=None):
        msg = self._alter_msg(msg, extra)
        stack = traceback.extract_stack(f=None, limit=2)
        filename = Path(stack[0].filename).name
        lineno = stack[0].lineno
        levelname = "WARN"
        super().warning(f"ln {lineno:4d}:{filename:25s}  {levelname:8s}  {msg}", *args, exc_info=exc_info, stack_info=stack_info,
                        stacklevel=stacklevel, extra=extra)

    def warn(self, msg, *args, exc_info=None, stack_info=False, stacklevel=1, extra=None):
        msg = self._alter_msg(msg, extra)
        stack = traceback.extract_stack(f=None, limit=2)
        filename = Path(stack[0].filename).name
        lineno = stack[0].lineno
        levelname = "WARN"
        super().warn(f"ln {lineno:4d}:{filename:25s}  {levelname:8s}  {msg}", *args, exc_info=exc_info, stack_info=stack_info, stacklevel=stacklevel,
                     extra=extra)

    def error(self, msg, *args, exc_info=None, stack_info=False, stacklevel=1, extra=None):
        msg = self._alter_msg(msg, extra)
        stack = traceback.extract_stack(f=None, limit=2)
        filename = Path(stack[0].filename).name
        lineno = stack[0].lineno
        levelname = "ERR"
        super().error(f"ln {lineno:4d}:{filename:25s}  {levelname:8s}  {msg}", *args, exc_info=exc_info, stack_info=stack_info, stacklevel=stacklevel,
                      extra=extra)

    def critical(self, msg, *args, exc_info=None, stack_info=False, stacklevel=1, extra=None):
        msg = self._alter_msg(msg, extra)
        stack = traceback.extract_stack(f=None, limit=2)
        filename = Path(stack[0].filename).name
        lineno = stack[0].lineno
        levelname = "CRIT"
        super().critical(f"ln {lineno:4d}:{filename:25s}  {levelname:8s}  {msg}", *args, exc_info=exc_info, stack_info=stack_info,
                         stacklevel=stacklevel, extra=extra)

    def exception(self, msg, *args, exc_info=True, stack_info=False, stacklevel=1, extra=None):
        msg = self._alter_msg(msg, extra)
        stack = traceback.extract_stack(f=None, limit=2)
        filename = Path(stack[0].filename).name
        lineno = stack[0].lineno
        levelname = "EXCE"
        super().exception(f"ln {lineno:4d}:{filename:25s}  {levelname:8s}  {msg}", *args, exc_info=exc_info, stack_info=stack_info,
                          stacklevel=stacklevel, extra=extra)





from typing import Optional

from PyQt6.QtWidgets import QWidget, QMessageBox, QDialogButtonBox


class YesOrNoNotificationWindow(QMessageBox):

    def __init__(self, message:str, title:str = "notification", alt_yes_text: Optional[str] = None, alt_no_text : Optional[str] = None):
        self.message = message
        self.title = title
        super().__init__()

        self.setWindowTitle(self.title)
        self.setText(self.message)
        self.setStandardButtons(QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
        if alt_yes_text:
            self.button(QMessageBox.StandardButton.Yes).setText(alt_yes_text)
        if alt_no_text:
            self.button(QMessageBox.StandardButton.No).setText(alt_no_text)


from collections import defaultdict
from pathlib import Path

from otlmow_converter.DotnotationDictConverter import DotnotationDictConverter
from otlmow_converter.Exceptions.CannotCombineAssetsError import CannotCombineAssetsError
from otlmow_converter.Exceptions.CannotCombineDifferentAssetsError import \
    CannotCombineAssetsWithDifferentIdError, CannotCombineAssetsWithDifferentTypeError
from otlmow_converter.Exceptions.ExceptionsGroup import ExceptionsGroup
from otlmow_converter.Exceptions.NoIdentificatorError import NoIdentificatorError
from otlmow_model.OtlmowModel.BaseClasses.OTLObject import OTLObject

from otlmow_gui.GUI.translation.GlobalTranslate import GlobalTranslate


class CombineAssetHelper:
    @classmethod
    def combine_assets(cls,asset_list: list[OTLObject], model_directory: Path = None) -> list[
        OTLObject]:
        assets = defaultdict(list)
        for asset in asset_list:
            if asset.typeURI == 'http://purl.org/dc/terms/Agent':
                assets[asset.agentId.identificator].append(asset)
            else:
                assets[asset.assetId.identificator].append(asset)
        combined_assets = []

        list_of_errors = []
        for asset_list in assets.values():

            try:
                if len(asset_list) == 1:
                    combined_assets.append(asset_list[0])
                else:
                    combined_asset = cls.combine_two_asset_instances(asset_list[0], asset_list[1],
                                                                 model_directory)
                    for asset in asset_list[2:]:
                        combined_asset = cls.combine_two_asset_instances(combined_asset, asset,
                                                                     model_directory)
                    combined_assets.append(combined_asset)

            except CannotCombineAssetsWithDifferentTypeError as ex:
                object_id = ex.object_id
                short_uri = ""
                short_uri_2 = ""
                if ex.attribute_errors and ex.attribute_errors[0][0] == 'typeURI':
                    short_uri = ex.attribute_errors[0][1][0].split('/')[-1]
                    short_uri_2 = ex.attribute_errors[0][1][1].split('/')[-1]

                ex.message = (GlobalTranslate._('Cannot combine the assets with id: "{object_id}"\n'
                              'due to conflicting types: {short_uri} != {short_uri_2}').format(object_id=object_id,short_uri=short_uri,short_uri_2=short_uri_2))

                ex.args = ex.message,
                list_of_errors.append(ex)
            except CannotCombineAssetsError as ex:
                object_id = ex.object_id
                short_uri = ex.type_uri.split('/')[-1]
                error_str = '\n'.join(
                    [f'{t[0]}: {t[1][0]} != {t[1][1]}' for t in ex.attribute_errors])

                ex.message = (
                        GlobalTranslate._('Cannot combine the assets with id: "{object_id}" with type "{short_uri}"\n'
                        'due to conflicting values in attribute(s):\n').format(object_id=object_id,short_uri=short_uri) + error_str)

                ex.args = ex.message,

                list_of_errors.append(ex)
        if list_of_errors:
            raise ExceptionsGroup(
                message= GlobalTranslate._('There were errors while combining the assets'),
                exceptions=list_of_errors
            )

        return combined_assets

    @classmethod
    def combine_two_asset_instances(cls,asset1: OTLObject, asset2: OTLObject,
                                    model_directory: Path = None) -> OTLObject:
        if asset1 is None or asset2 is None:
            raise ValueError('One of the assets is None')

        id1 = asset1.assetId.identificator if asset1.typeURI != 'http://purl.org/dc/terms/Agent' \
            else asset1.agentId.identificator
        id2 = asset2.assetId.identificator if asset2.typeURI != 'http://purl.org/dc/terms/Agent' \
            else asset2.agentId.identificator

        if id1 is None or id2 is None:
            raise NoIdentificatorError('One of the assets has no assetId.identificator')

        if id1 != id2:
            ex = CannotCombineAssetsWithDifferentIdError(
                'The assets have different identificator values')
            ex.attribute_errors = [('identificator', (id1, id2))]
            ex.type_uri = asset1.typeURI
            raise ex

        if asset1.typeURI != asset2.typeURI:
            ex = CannotCombineAssetsWithDifferentTypeError('The assets have different types')
            ex.attribute_errors = [('typeURI', (asset1.typeURI, asset2.typeURI))]
            ex.object_id = id1
            raise ex

        ddict1 = DotnotationDictConverter.to_dict(asset1)
        ddict2 = DotnotationDictConverter.to_dict(asset2)
        if asset2.typeURI == 'http://purl.org/dc/terms/Agent':
            ddict2.pop('agentId.identificator')
        else:
            ddict2.pop('assetId.identificator')

        attribute_errors = []
        for key, value in ddict2.items():
            if value is None:
                continue
            if key == 'typeURI':
                continue

            if key in ddict1 and ddict1[key] is not None and ddict1[key] != value:
                attribute_errors.append((key, ddict1[key], value))
            else:
                ddict1[key] = value

        if attribute_errors:
            error_str = '\n'.join(
                [f'{key}: {ddict1[key]}, {value}' for key, _, value in sorted(attribute_errors)])
            ex = CannotCombineAssetsError(
                message=f'Cannot combine the assets with id {id1} because some attributes '
                        'have conflicting values:\n'
                        f'{error_str}')
            ex.object_id = id1
            ex.type_uri = asset1.typeURI
            ex.attribute_errors = [(key, (ddict1[key], value)) for key, _, value in
                                   attribute_errors]
            raise ex

        return DotnotationDictConverter.from_dict(ddict1, model_directory=model_directory)


import asyncio
import datetime
import xml.etree.cElementTree as ET
from pathlib import Path
from typing import Optional
from xml.etree.ElementTree import Element

from otlmow_converter.DotnotationHelper import DotnotationHelper
from otlmow_model.OtlmowModel.BaseClasses.OTLObject import dynamic_create_instance_from_uri, OTLAttribuut, OTLObject
from otlmow_model.OtlmowModel.Helpers.OTLObjectHelper import is_relation
from otlmow_modelbuilder.OSLOCollector import OSLOCollector

from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception


class XSDCreator:

    @classmethod
    def create_xsd_from_subset(cls, subset_path: Path, xsd_path: Path,
                               model_directory: Path = None) -> None:
        create_task_reraise_exception(cls.create_filtered_xsd_from_subset(subset_path=subset_path, xsd_path=xsd_path,
                                            selected_classes_typeURI_list=[], model_directory=model_directory))

    @classmethod
    async def create_filtered_xsd_from_subset(
            cls, subset_path: Path, xsd_path: Path, selected_classes_typeURI_list: Optional[list[str]]=None,
            model_directory: Path = None) -> None:

        collector = OSLOCollector(subset_path)
        collector.collect_all(include_abstract=True)

        ET.register_namespace("xs", 'http://www.w3.org/2001/XMLSchema')
        ET.register_namespace("gml", 'http://www.opengis.net/gml')
        ET.register_namespace("Schema1", 'http://fdo.osgeo.org/schemas/feature/Schema1')
        ET.register_namespace("fdo", 'http://fdo.osgeo.org/schemas')

        schema = ET.Element("{http://www.w3.org/2001/XMLSchema}schema", targetNamespace="http://fdo.osgeo.org/schemas/feature/Schema1",  elementFormDefault="qualified", attributeFormDefault="unqualified" )
        annotation = ET.SubElement(schema, "{http://www.w3.org/2001/XMLSchema}annotation")
        documentation = ET.SubElement(annotation, "{http://www.w3.org/2001/XMLSchema}documentation")
        documentation.text = "Default schema"
        tree = ET.ElementTree(schema)

        if selected_classes_typeURI_list:
            classes = [class_ for class_ in  collector.classes if class_.objectUri in selected_classes_typeURI_list]
        else:
            # if no selected_classes are given take everything from subset
            classes = collector.classes

        for class_ in classes:
            if class_.abstract:
                continue
            class_instance = dynamic_create_instance_from_uri(class_.objectUri, model_directory=model_directory)
            if is_relation(class_instance):
                continue

            class_element = ET.SubElement(
                schema, "{http://www.w3.org/2001/XMLSchema}element",
                attrib={'name': f"OTL_{class_.name}",
                        'type': f"http://fdo.osgeo.org/schemas/feature/Schema1/OTL_{class_.name}Type",
                        'abstract': "false", 'substitutionGroup': "http://www.opengis.net/gml/_Feature"})
            key = ET.SubElement(class_element, "{http://www.w3.org/2001/XMLSchema}key", name=f"OTL_{class_.name}Key")
            ET.SubElement(key, "{http://www.w3.org/2001/XMLSchema}selector", xpath=f".//OTL_{class_.name}")
            ET.SubElement(key, "{http://www.w3.org/2001/XMLSchema}field", xpath="FeatId")



            complex_type = ET.SubElement(schema, "{http://www.w3.org/2001/XMLSchema}complexType",
                                         attrib={'name': f"OTL_{class_.name}Type", 'abstract': "false",
                                                 '{http://fdo.osgeo.org/schemas}geometryName': "Geometry"})
            annotation = ET.SubElement(complex_type, "{http://www.w3.org/2001/XMLSchema}annotation")
            documentation = ET.SubElement(annotation, "{http://www.w3.org/2001/XMLSchema}documentation")
            documentation.text = class_instance.__doc__
            if documentation.text is not None:
                documentation.text = documentation.text[:159]
            complex_content = ET.SubElement(complex_type, "{http://www.w3.org/2001/XMLSchema}complexContent")
            extension = ET.SubElement(complex_content, "{http://www.w3.org/2001/XMLSchema}extension",
                                      attrib={'base': "http://www.opengis.net/gml/AbstractFeatureType"})
            attribute_sequence = ET.SubElement(extension, "{http://www.w3.org/2001/XMLSchema}sequence")

            cls.add_feat_id_in_attribute_sequence(attribute_sequence)

            cls.add_geometry_in_attribute_sequence(attribute_sequence, class_instance)

            attribute_names_in_subset = [attribuut.name for attribuut in collector.find_attributes_by_class(class_)]

            if 'geometry' in attribute_names_in_subset:
                attribute_names_in_subset.remove('geometry')
            class_instance.fill_with_dummy_data()
            DotnotationHelper.clear_list_of_list_attributes(class_instance)

            cls.add_otl_attributes_in_attribute_sequence(attribute_names_in_subset, attribute_sequence, class_instance)
            await asyncio.sleep(0)

        tree.write(xsd_path, encoding='utf-8', xml_declaration=True, method='xml')

        cls._tweak_file(xsd_path)

    @classmethod
    def add_geometry_in_attribute_sequence(cls, attribute_sequence: Element, class_instance: OTLObject) -> None:
        if not hasattr(class_instance, '_geometry_types'):
            geo_types = ""
        else:
            geo_types = ' '.join(class_instance._geometry_types).replace(' Z', '').lower()
            if geo_types != '':
                geo_types += ' '

        geometry = ET.SubElement(
            attribute_sequence, "{http://www.w3.org/2001/XMLSchema}element",
            attrib={'name': "Geometry", 'type': "http://www.opengis.net/gml/AbstractGeometryType",
                    '{http://fdo.osgeo.org/schemas}hasMeasure': "false",
                    '{http://fdo.osgeo.org/schemas}hasElevation': "true",
                    '{http://fdo.osgeo.org/schemas}srsName': "Default",
                    '{http://fdo.osgeo.org/schemas}geometricTypes': geo_types,
                    '{http://fdo.osgeo.org/schemas}geometryTypes': geo_types}
        )
        annotation = ET.SubElement(geometry, "{http://www.w3.org/2001/XMLSchema}annotation")
        documentation = ET.SubElement(annotation, "{http://www.w3.org/2001/XMLSchema}documentation")
        documentation.text = "Default geometry property"

    @classmethod
    def add_feat_id_in_attribute_sequence(cls, attribute_sequence: Element):
        feat_id = ET.SubElement(attribute_sequence, "{http://www.w3.org/2001/XMLSchema}element",
                                attrib={'name': "FeatId", '{http://fdo.osgeo.org/schemas}readOnly': "true",
                                        '{http://fdo.osgeo.org/schemas}autogenerated': "true"})
        annotation = ET.SubElement(feat_id, "{http://www.w3.org/2001/XMLSchema}annotation")
        documentation = ET.SubElement(annotation, "{http://www.w3.org/2001/XMLSchema}documentation")
        documentation.text = "Default identity property"
        simple_type = ET.SubElement(feat_id, "{http://www.w3.org/2001/XMLSchema}simpleType")
        ET.SubElement(simple_type, "{http://www.w3.org/2001/XMLSchema}restriction", base="xs:int")

    @classmethod
    def add_otl_attributes_in_attribute_sequence(cls, attribute_names_in_subset: [str], attribute_sequence: Element,
                                                 class_instance: OTLObject) -> None:
        for dotnotation, _ in DotnotationHelper.list_attributes_and_values_by_dotnotation(
                class_instance, waarde_shortcut=True):
            first_in_dotnotation = dotnotation
            if '.' in first_in_dotnotation:
                first_in_dotnotation = first_in_dotnotation.split('.')[0]
            if '[]' in first_in_dotnotation:
                first_in_dotnotation = first_in_dotnotation[:-2]
            if first_in_dotnotation not in attribute_names_in_subset:
                continue

            attr_instance = DotnotationHelper.get_attribute_by_dotnotation(class_instance, dotnotation)
            definition = attr_instance.definition
            if hasattr(attr_instance.owner, '_parent'):
                parent_attribute_for_waarde_shortcut = attr_instance.owner._parent
                if parent_attribute_for_waarde_shortcut.field.waarde_shortcut_applicable:
                    dotnotation = DotnotationHelper.get_dotnotation(parent_attribute_for_waarde_shortcut)
                    definition = parent_attribute_for_waarde_shortcut.definition

            element = ET.SubElement(attribute_sequence, "{http://www.w3.org/2001/XMLSchema}element",
                                    name=dotnotation.replace('.', '_'), minOccurs="0", default="")
            annotation = ET.SubElement(element, "{http://www.w3.org/2001/XMLSchema}annotation")
            documentation = ET.SubElement(annotation, "{http://www.w3.org/2001/XMLSchema}documentation")
            documentation.text = definition
            if documentation.text is not None:
                documentation.text = documentation.text[:159]

            cls.create_type_from_attribute_in_element(element, attr_instance)

        # hard add typeURI
        element = ET.SubElement(attribute_sequence, "{http://www.w3.org/2001/XMLSchema}element",
                                name='typeURI', minOccurs="0", default=class_instance.typeURI)
        annotation = ET.SubElement(element, "{http://www.w3.org/2001/XMLSchema}annotation")
        documentation = ET.SubElement(annotation, "{http://www.w3.org/2001/XMLSchema}documentation")
        documentation.text = 'De URI van het object volgens https://www.w3.org/2001/XMLSchema#anyURI.'
        simple_type = ET.SubElement(element, "{http://www.w3.org/2001/XMLSchema}simpleType")
        restriction = ET.SubElement(simple_type, "{http://www.w3.org/2001/XMLSchema}restriction", base="xs:string")
        ET.SubElement(restriction, "{http://www.w3.org/2001/XMLSchema}maxLength", value="9999")

    @classmethod
    def _tweak_file(cls, xsd_path):
        # open file
        with open(xsd_path, 'r') as file:
            data = file.readlines()

        new_data = []
        for line in data:
            line = line.replace('>', '>\n')
            new_data.extend(line.split('\n'))

        # tweak file
        if new_data[3][:10] == '<xs:schema':
            new_data[3] = ('<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema" targetNamespace="http://fdo.osgeo.org/schemas/feature/Schema1" '
                           'xmlns:fdo="http://fdo.osgeo.org/schemas" xmlns:gml="http://www.opengis.net/gml" '
                           'xmlns:Schema1="http://fdo.osgeo.org/schemas/feature/Schema1" elementFormDefault="qualified" '
                           'attributeFormDefault="unqualified">')

        for index, line in enumerate(new_data):
            if line.startswith('<xs:element name='):
                new_data[index] = line.replace('http://fdo.osgeo.org/schemas/feature/Schema1/', 'Schema1:').replace('http://www.opengis.net/gml/', 'gml:')
            elif line.startswith('<xs:extension base="http://www.opengis.net/gml/AbstractFeatureType">'):
                new_data[index] = '<xs:extension base="gml:AbstractFeatureType">'

        # write file
        with open(xsd_path, 'w') as file:
            file.writelines(new_data)

    @classmethod
    def create_type_from_attribute_in_element(cls, element: Element, attr_instance: OTLAttribuut):
        if attr_instance.field.native_type == float:
            simple_type = ET.SubElement(element, "{http://www.w3.org/2001/XMLSchema}simpleType")
            restriction = ET.SubElement(simple_type, "{http://www.w3.org/2001/XMLSchema}restriction", base="xs:decimal")
            ET.SubElement(restriction, "{http://www.w3.org/2001/XMLSchema}totalDigits", value="19")
            ET.SubElement(restriction, "{http://www.w3.org/2001/XMLSchema}fractionDigits", value="10")
        elif attr_instance.field.native_type == bool:
            simple_type = ET.SubElement(element, "{http://www.w3.org/2001/XMLSchema}simpleType")
            ET.SubElement(simple_type, "{http://www.w3.org/2001/XMLSchema}restriction", base="xs:boolean")
        elif attr_instance.field.native_type == int:
            simple_type = ET.SubElement(element, "{http://www.w3.org/2001/XMLSchema}simpleType")
            ET.SubElement(simple_type, "{http://www.w3.org/2001/XMLSchema}restriction", base="xs:int")
        elif hasattr(attr_instance.field, 'options'):
            simple_type = ET.SubElement(element, "{http://www.w3.org/2001/XMLSchema}simpleType")
            restriction = ET.SubElement(simple_type, "{http://www.w3.org/2001/XMLSchema}restriction", base="xs:string")
            ET.SubElement(restriction, "{http://www.w3.org/2001/XMLSchema}enumeration", value='-')
            element.attrib.update({'default': '-'})
            max_length = 50
            for keuze, keuzelijstwaarde in attr_instance.field.options.items():
                if keuzelijstwaarde.status != 'ingebruik':
                    continue
                ET.SubElement(restriction, "{http://www.w3.org/2001/XMLSchema}enumeration", value=keuze)
                max_length = max(max_length, len(keuze))
            ET.SubElement(restriction, "{http://www.w3.org/2001/XMLSchema}maxLength", value=str(max_length))
        elif attr_instance.field.native_type in {str, datetime.date, datetime.datetime, datetime.time} or attr_instance.field.objectUri == 'http://www.w3.org/2001/XMLSchema#string':
            simple_type = ET.SubElement(element, "{http://www.w3.org/2001/XMLSchema}simpleType")
            restriction = ET.SubElement(simple_type, "{http://www.w3.org/2001/XMLSchema}restriction", base="xs:string")
            ET.SubElement(restriction, "{http://www.w3.org/2001/XMLSchema}maxLength", value="9999")
        else:
            raise NotImplementedError(f"Field {attr_instance.field} not implemented")



from pathlib import Path

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QDialogButtonBox, QDialog, QVBoxLayout, QFrame, QHBoxLayout, QLabel, QLineEdit, QPushButton
import qtawesome as qta

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.step_domain.HomeDomain import HomeDomain
from otlmow_gui.Exceptions.WrongDatabaseError import WrongDatabaseError
from otlmow_gui.GUI.dialog_windows.file_picker_dialog.SubsetLoadFilePickerDialog import \
    SubsetLoadFilePickerDialog


class ChangeSubsetWindow(QDialog):

    def __init__(self, language_settings):
        super().__init__()
        self._ = language_settings

        project = global_vars.current_project

        self.setModal(True)
        self.setMinimumWidth(700)
        self.setWindowTitle(self._("change_subset"))

        layout = QVBoxLayout()

        frame = QFrame()
        horizontal_layout = QHBoxLayout()

        self.subset_label = QLabel(self._("subset") + ":")

        self.input_subset = QLineEdit()
        self.input_subset.setReadOnly(True)
        self.input_subset.setText(str(project.subset_path))

        file_picker_btn = QPushButton()
        file_picker_btn.setIcon(qta.icon('mdi.folder-open-outline'))
        file_picker_btn.clicked.connect(lambda: self.open_file_picker(self.input_subset))

        horizontal_layout.addWidget(self.subset_label, alignment=Qt.AlignmentFlag.AlignLeft)
        horizontal_layout.addWidget(self.input_subset)
        horizontal_layout.addWidget(file_picker_btn)

        frame.setLayout(horizontal_layout)
        layout.addWidget(frame)

        self.error_label = QLabel()
        self.error_label.setText("")
        self.error_label.setStyleSheet("color: red")

        button_box = self.create_button_box()
        button_box.accepted.connect(
            lambda: self.validate_change_subset(input_subset=self.input_subset.text(), old_project_path=project.project_path))
        button_box.rejected.connect(self.reject)

        layout.addWidget(self.error_label)
        layout.addWidget(button_box)
        self.setLayout(layout)

        self.show()
        self.exec()

    def validate_change_subset(self, input_subset: str, old_project_path: Path) -> None:
        """
        Validates and applies a change to the current subset.

        This function attempts to change the subset to the specified input path. If the change is
        unsuccessful due to a WrongDatabaseError, it updates the error label with the error message
        and resets the input field to the old project path.

        Args:
            self: The instance of the class.
            input_subset (str): The new subset path to be validated and applied.
            old_project_path (Path): The path of the old project to revert to in case of an error.

        Returns:
            None
        """

        try:
            HomeDomain.change_subset(new_path=input_subset)
        except WrongDatabaseError as e:
            self.error_label.setText(str(e))
            self.input_subset.setText(str(old_project_path))
            return
        except FileNotFoundError as e:
            self.error_label.setText(f"Can't find subset: {str(Path(input_subset).name)}")
            self.input_subset.setText(str(old_project_path))
            return
        self.close()

    def create_button_box(self) -> QDialogButtonBox:
        """
        Creates a button box for the dialog window.

        This function initializes a QDialogButtonBox with "Ok" and "Cancel" buttons, setting their
        properties and text labels. It provides a user-friendly interface for submitting or
        canceling actions within the dialog.

        Args:
            self: The instance of the class.

        Returns:
            QDialogButtonBox: The configured button box containing the action buttons.
        """

        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        button_box.setProperty("class", "button-box")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setProperty("class", "primary-button")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setText(self._("submit"))
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setProperty("class", "secondary-button")
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setText(self._("cancel"))

        return button_box

    @classmethod
    def open_file_picker(cls, input_subset:QLineEdit) -> None:
        """
        Opens a file picker dialog to select a subset file.

        This class method retrieves the starting directory for the file picker based on the
        current input and allows the user to select a database file. It updates the input field
        with the selected file path and stores the directory for future use.

        Args:
            cls: The class itself.
            input_subset: The input field where the selected file path will be set.

        Returns:
            None
        """

        file_path = global_vars.get_start_dir_subset_selection(
            input_subset_str= input_subset.text())

        selected_list = SubsetLoadFilePickerDialog.instance.summon(directory=file_path)
        cls.set_input_line_to_new_subset_path(input_subset, selected_list)

    @classmethod
    def set_input_line_to_new_subset_path(cls, input_subset:QLineEdit, selected:list[Path]):
        if not selected:
            return

        selected_file:Path= selected[0]
        input_subset.setText(str(selected_file))
        if selected_file:
            global_vars.last_subset_selected_dir = selected_file




from PyQt6.QtWidgets import QDialog, QVBoxLayout, QLabel, QDialogButtonBox

from otlmow_gui.Domain.project.Project import Project


class RemoveProjectFilesWindow:

    def __init__(self, project: Project, language_settings):
        self.project = project
        self._ = language_settings
        self.init_ui()

    def init_ui(self):
        dialog = QDialog()
        dialog.setModal(True)
        dialog.setWindowTitle(self._("remove_project_files_title"))
        layout = QVBoxLayout()
        question_label = QLabel()
        question_label.setText(self._("remove_project_files_question"))
        layout.addWidget(question_label)
        button_box = self.create_button_box()
        button_box.accepted.connect(lambda: self.remove_project_files(dialog))
        button_box.rejected.connect(dialog.reject)
        layout.addWidget(button_box)
        dialog.setLayout(layout)
        dialog.show()
        dialog.exec()

    def create_button_box(self):
        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        button_box.setProperty("class", "button-box")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setProperty("class", "primary-button")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setText(self._("submit"))
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setProperty("class", "secondary-button")
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setText(self._("cancel"))
        return button_box

    def remove_project_files(self, dialog):
        self.project.remove_all_project_files()
        dialog.close()




from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QDialog, QLabel, QVBoxLayout, QHBoxLayout, QLineEdit, QDialogButtonBox, \
    QComboBox

from otlmow_gui.Domain.util.Helpers import Helpers

from otlmow_gui.Domain.step_domain.RelationChangeDomain import RelationChangeDomain
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception


class AddExternalAssetWindow:


    def __init__(self, language_settings):
        self.error_label = QLabel()
        self._ = language_settings

        self.input_asset_id_or_name: QLineEdit = QLineEdit()
        self.combobox_asset_type: QComboBox = QComboBox()



    def draw_add_external_asset_window(self):

        self.error_label.setText("")
        dialog_window = QDialog()
        dialog_window.setMinimumWidth(450)
        # Makes the dialog the primary screen, disabling the screen behind it
        dialog_window.setModal(True)
        dialog_window.setWindowTitle(self._("add_external_asset_title"))
        # Creates the vertical stack layout
        layout = QVBoxLayout()

        # Creates 3 horizontal layouts for each input field with its label
        container_eigen_ref = QHBoxLayout()
        label_asset_id_or_name = QLabel( self._("asset_id_or_name")+ ":")
        self.input_asset_id_or_name = QLineEdit()
        self.input_asset_id_or_name.setPlaceholderText(self._("asset_id_dummy"))
        container_eigen_ref.addWidget(label_asset_id_or_name, alignment=Qt.AlignmentFlag.AlignLeft)
        container_eigen_ref.addWidget(self.input_asset_id_or_name)

        container_subset = QHBoxLayout()
        label_asset_type = QLabel(self._("asset_type") + ":")

        self.combobox_asset_type = self.create_combobox(Helpers.all_OTL_asset_types_dict)
        self.combobox_asset_type.setPlaceholderText(self._("asset_type_dummy"))
        container_subset.addWidget(label_asset_type, alignment=Qt.AlignmentFlag.AlignLeft)
        container_subset.addWidget(self.combobox_asset_type)

        # Adds the input fields to the layout
        layout.addLayout(container_eigen_ref)
        layout.addLayout(container_subset)

        # Changes the color of the error label to red
        self.error_label.setStyleSheet("color: red")

        # Creates the button box
        button_box:QDialogButtonBox = self.create_button_box()
        # sends the values off to validate once submitted
        button_box.accepted.connect(lambda: self.add_asset(dialog_window))
        button_box.rejected.connect(dialog_window.reject)
        # Adds the two buttons to the layout
        layout.addWidget(button_box)
        layout.addWidget(self.error_label)
        # Fills the dialog with the created layout
        dialog_window.setLayout(layout)
        # Shows the dialog
        dialog_window.show()
        dialog_window.exec()

    def add_asset(self, dialog_window):
        id_or_name = self.input_asset_id_or_name.text().strip()
        if not id_or_name:
            error_text = self._("asset_id_or_name") + self._(" is leeg")
            self.error_label.setText(str(error_text))
            return

        if RelationChangeDomain.get_object(id_or_name):
            error_text = self._('OTL-asset with {0} "{1}" already exists'.format(self._("asset_id_or_name"),id_or_name ))
            self.error_label.setText(str(error_text))
            return

        combobox_choice = self.combobox_asset_type.currentText()

        type_uri = Helpers.all_OTL_asset_types_dict[combobox_choice]

        # Replaced by an async call
        # RelationChangeDomain.create_and_add_new_external_asset(id_or_name=id_or_name, type_uri=type_uri)
        # global_vars.current_project.visualisation_uptodate.set_clear_all(True)
        # RelationChangeDomain.update_frontend()

        #async version that also saves the assets after executing
        create_task_reraise_exception(RelationChangeDomain.user_input_to_create_and_add_new_external_asset(id_or_name=id_or_name, type_uri=type_uri))

        dialog_window.close()

    def create_button_box(self):
        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        button_box.setProperty("class", "button-box")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setProperty("class", "primary-button")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setText(self._("submit"))
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setProperty("class", "secondary-button")
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setText(self._("cancel"))
        return button_box

    def create_combobox(self, options_to_data_dict):
        comboBox = QComboBox()
        comboBox.addItems(list(options_to_data_dict.keys()))
        # comboBox.currentTextChanged.connect(self.)

        comboBox.setEditable(True)
        comboBox.view().setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)

        return comboBox



class WrongFileTypeError(Exception):

    def __init__(self, language, expected_filetype_name,expected_filetype_suffix):
        self._ = language
        self.expected_filetype_name = expected_filetype_name
        self.expected_filetype_suffix = expected_filetype_suffix

        super().__init__()


    def __str__(self):
        translation = self._("The path to the provided file is not a {0} file with extension ({1})")
        return translation.format(self.expected_filetype_name,self.expected_filetype_suffix)



from pathlib import Path
from typing import Optional

from PyQt6.QtWidgets import QFileDialog

from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.GUI.dialog_windows.file_picker_dialog.AbstractFilePickerDialog import AbstractFilePickerDialog


class SaveFilePickerDialog(AbstractFilePickerDialog):
    def __init__(self, language_settings,title:Optional[str] = None,name_filter:str=None, initial_file_path: Optional[str]=None,
                 exclude_file_types:list[str]=[]):


        super().__init__(language_settings, title, name_filter, initial_file_path)
        if not title:
            self.setWindowTitle(self._("Exporteer bestand(en)"))  # standard title for save dialogs

        if not name_filter:
            for excluded_type in exclude_file_types:
                if excluded_type in self.supported_export_formats:
                    self.supported_export_formats.pop(
                        excluded_type)  # SDF is not yet supported for export in V1.0
        self.setModal(True)
        self.setAcceptMode(QFileDialog.AcceptSave)


    def execute(self):
        if not self.exec():
            return[]

        save_location = self.selectedFiles()

        if not save_location:
            OTLLogger.logger.debug(f"save_file_dialog return: {[]}")
            return []
        if not save_location[0]:
            OTLLogger.logger.debug(f"save_file_dialog return: {[]}")
            return []

        OTLLogger.logger.debug(f"save_file_dialog return: {[Path(save_location[0])]}")
        return [Path(save_location[0])]

    def store_save_filename(self, res: list[str]):
        if len(res):
            if res[0]:
                self.previous_exported_file_name = str(Path(res[0]).stem)

    def set_filename_suggestions(self, project_name):

        if self.previous_exported_file_name:
            initial_filename = self.previous_exported_file_name
        elif project_name:
            initial_filename = self.get_filename_suggestion(project_name)
        else:
            initial_filename = "export"

        self.selectFile(initial_filename)

    def get_filename_suggestion(self, project_name):
        return project_name + "_export"








from dataclasses import dataclass
from pathlib import Path
from typing import List

from otlmow_converter.Exceptions.ExceptionsGroup import ExceptionsGroup
from otlmow_model.OtlmowModel.BaseClasses.OTLObject import create_dict_from_asset, OTLObject
from otlmow_model.OtlmowModel.Helpers.OTLObjectHelper import \
    compare_two_lists_of_objects_attribute_level, is_relation
from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.enums import ReportAction
from otlmow_gui.Domain.step_domain.ExportDataDomain import ExportDataDomain
from otlmow_gui.Domain.step_domain.InsertDataDomain import InsertDataDomain
from otlmow_gui.Domain.step_domain.RelationChangeDomain import async_save_assets, RelationChangeDomain
from otlmow_gui.GUI.dialog_windows.LoadingImageWindow import add_loading_screen, add_loading_screen_no_delay
from otlmow_gui.GUI.dialog_windows.NotificationWindow import NotificationWindow
from otlmow_gui.GUI.translation.GlobalTranslate import GlobalTranslate


@dataclass
class ReportItem:
    id: str
    actie: ReportAction
    attribute: str
    original_value: str
    new_value: str

class ExportFilteredDataSubDomain:
    original_documents:dict[str,Path] = {}


    @classmethod
    def init_static(cls):
        cls.clear_data()

    @classmethod
    def clear_data(cls):
        cls.original_documents = {}

    @classmethod
    def add_original_documents(cls,paths:list[Path]):
        for path in paths:
            filename = path.name
            cls.original_documents[filename] = path
        cls.update_frontend()

    @classmethod
    def delete_original_file(cls,doc_name:str):
        cls.original_documents.pop(doc_name)
        cls.update_frontend()

    @classmethod
    def update_frontend(cls):
        cls.get_screen().update_original_files_list(cls.original_documents)

    @classmethod
    def generate_diff_report(cls, original_data: list, new_data: list, model_directory: Path=None) -> List[ReportItem]:
        report_list = []
        diff_lists = cls.generate_difference_between_two_lists(list1=original_data, list2=new_data,
                                                               model_directory=model_directory)

        original_data_dict = {item.assetId.identificator: item for item in original_data}
        for item in diff_lists:
            old_item = original_data_dict.get(item.assetId.identificator)
            if old_item is None:
                if is_relation(item):
                    report_list.append(cls.generate_new_relation_report(item))
                else:
                    report_list.append(cls.generate_new_asset_report(item))
            else:
                report_list.extend(cls.generate_asset_change_report(item, old_item))
        return report_list

    @classmethod
    def generate_asset_change_report(cls, item, old_item) -> List[ReportItem]:
        report_list = []
        item_dict = create_dict_from_asset(item)
        old_item_dict = create_dict_from_asset(old_item)
        for key, value in item_dict.items():
            if key not in ['assetId', 'typeURI']:
                if isinstance(value, dict):
                    report_list.extend(
                        cls.generate_complex_asset_report(item=item, attribute=key, old_item_dict=old_item_dict,
                                                          complex_list=value))
                else:
                    report_list.append(cls.generate_simple_asset_report(item=item, key=key, value=value,
                                                                        old_item_dict=old_item_dict))
        return report_list

    @classmethod
    @add_loading_screen_no_delay
    @async_save_assets
    async def get_diff_report(cls) -> None:
        # model_dir = ProgramFileStructure.get_otl_wizard_model_dir()
        original_assets, error_set = await cls.generate_original_assets_from_files(original_documents=list(cls.original_documents.values()))

        if error_set:
            cls.get_screen().negative_feedback_message()
            cls.get_screen().fill_up_change_table_with_error_feedback(error_set)
            return

        new_assets = RelationChangeDomain.get_export_instances()

        cls.get_screen().positive_feedback_message()
        cls.get_screen().fill_up_change_table(cls.generate_diff_report(original_assets, new_assets))

    @classmethod
    def sync_export_diff_report(cls,
                           file_name: str,
                           separate_per_class_csv_option:bool = False,
                           separate_relations_option:bool = False,**kwargs) -> None:
        cls.export_diff_report(file_name= file_name,
                               separate_per_class_csv_option=separate_per_class_csv_option,
                               separate_relations_option=separate_relations_option,**kwargs)

    @classmethod
    @add_loading_screen
    async def export_diff_report(cls,
                           file_name: str,
                           separate_per_class_csv_option:bool = False,
                           separate_relations_option:bool = False,**kwargs) -> None:
        OTLLogger.logger.debug("started exporting diff report")
        try:

            changed_assets = sorted(RelationChangeDomain.get_internal_objects(),
                                      key=lambda relation1: relation1.typeURI)
            changed_relations = sorted(RelationChangeDomain.get_persistent_relations(),
                                         key=lambda relation1: relation1.typeURI)

            original_objects, error_set = await cls.generate_original_assets_from_files(original_documents=list(cls.original_documents.values()))
            original_assets = [original_object for original_object in original_objects if not is_relation(original_object)]
            original_relations =[original_object for original_object in original_objects if is_relation(original_object)]



            diff_1_assets = compare_two_lists_of_objects_attribute_level(first_list=original_assets,
                                                                  second_list=changed_assets)
            diff_1_relations = compare_two_lists_of_objects_attribute_level(first_list=original_relations,
                                                                         second_list=changed_relations)

            assets = sorted(diff_1_assets,key=lambda relation1: relation1.typeURI)
            relations = sorted(diff_1_relations, key=lambda relation1: relation1.typeURI)

            await ExportDataDomain.export_to_files(assets, relations , file_name,
                                separate_per_class_csv_option, separate_relations_option,**kwargs)
            cls.get_screen().open_folder_of_created_export_files(file_name)
        except ValueError as e:
            if e.args == ('There are no asset data to export to Excel',):
                notification = NotificationWindow(title=GlobalTranslate._("Geen data"),
                                                  message=GlobalTranslate._("Er is geen data om te exporteren"))
                notification.exec()
            else:
                raise e
        except Exception as e:
            notification = NotificationWindow(title=GlobalTranslate._("Fout in originele bestanden"),
                                              message=GlobalTranslate._("Er is een fout in de originele bestanden\nKlik op \"Toon verschillen\" om een overzicht van de fouten te zien"))
            notification.exec()
            return



    @classmethod
    async def generate_original_assets_from_files(cls,original_documents: List[Path]) -> tuple[List[OTLObject],list]:
        original_assets = []
        error_set = []
        for path in original_documents:
            assets, exceptions_group = await InsertDataDomain.check_document( doc_location=path)

            if exceptions_group and exceptions_group.exceptions:
                for ex in exceptions_group.exceptions:
                    error_set.append({"exception": ex, "path_str": path.name})

            rounded_attribute_assets = []
            for asset in assets:
                waarde_shortcut = True
                allow_non_otl_conform_attributes = True
                warn_for_non_otl_conform_attributes = True

                d = create_dict_from_asset(asset, cast_datetime=True,
                                           waarde_shortcut=waarde_shortcut,
                                           allow_non_otl_conform_attributes=allow_non_otl_conform_attributes,
                                           warn_for_non_otl_conform_attributes=warn_for_non_otl_conform_attributes)

                rounded_asset = OTLObject.from_dict(d,
                                    waarde_shortcut=waarde_shortcut,
                                    cast_datetime=True,
                                    allow_non_otl_conform_attributes=allow_non_otl_conform_attributes,
                                    warn_for_non_otl_conform_attributes=warn_for_non_otl_conform_attributes)

                rounded_attribute_assets.append(rounded_asset)

            original_assets.extend(rounded_attribute_assets)
        try:
            original_assets = await InsertDataDomain.combine_assets_wrapper(original_assets)
        except ExceptionsGroup as group:
            if group.exceptions:
                for ex in group.exceptions:
                    error_set.append({"exception": ex, "path_str": ""})

        return original_assets, error_set

    @classmethod
    def generate_new_asset_report(cls,item) -> ReportItem:
        return ReportItem(
            id=item.assetId.identificator,
            actie=ReportAction.ASS,
            attribute="",
            original_value="",
            new_value=""
        )

    @classmethod
    def generate_new_relation_report(cls, item) -> ReportItem:
        return ReportItem(
            id=item.assetId.identificator,
            actie=ReportAction.REL,
            attribute="",
            original_value="",
            new_value=""
        )

    @classmethod
    def generate_complex_asset_report(cls,item, attribute, complex_list, old_item_dict) -> List[ReportItem]:
        return [ReportItem(
            id=item.assetId.identificator, actie=ReportAction.ATC, attribute=attribute,
            original_value="{0}: {1}".format(str(key),str(old_item_dict.get(attribute).get(key))),
            new_value="{0}: {1}".format(str(key),str(value))
        ) for key, value in complex_list.items()]

    @classmethod
    def generate_simple_asset_report(cls,item, key, value, old_item_dict) -> ReportItem:
        return ReportItem(
            id=item.assetId.identificator,
            actie=ReportAction.ATC,
            attribute=str(key),
            original_value=str(old_item_dict.get(key)),
            new_value=str(value)
        )

    @classmethod
    def generate_difference_between_two_lists(cls,list1: list, list2: list, model_directory: Path=None) -> list:
        diff_1 = compare_two_lists_of_objects_attribute_level(
            first_list=list1, second_list=list2, model_directory=model_directory)
        return compare_two_lists_of_objects_attribute_level(
            first_list=list1, second_list=diff_1, model_directory=model_directory)

    @classmethod
    def get_screen(cls):
        return global_vars.otl_wizard.main_window.step4_export.sub_screen_option_2_only_unedited_data

    @classmethod
    def remove_all_original_documents(cls):
        cls.original_documents.clear()

        cls.update_frontend()
